# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2010-2016 Red Hat, Inc.
#
# Authors:
# Thomas Woerner <twoerner@redhat.com>

import os

import dbus
import dbus.service
import collections

from firewall import config
from firewall.core.base import DEFAULT_ZONE_TARGET
from firewall.core.watcher import Watcher
from firewall.core.logger import log
from firewall.server.dbus import DbusServiceObject
from firewall.server.decorators import (
    handle_exceptions,
    dbus_handle_exceptions,
    dbus_service_method,
    dbus_service_signal,
    dbus_service_method_deprecated,
    dbus_service_signal_deprecated,
    dbus_polkit_require_auth,
)
from firewall.server.config_icmptype import FirewallDConfigIcmpType
from firewall.server.config_service import FirewallDConfigService
from firewall.server.config_zone import FirewallDConfigZone
from firewall.server.config_policy import FirewallDConfigPolicy
from firewall.server.config_ipset import FirewallDConfigIPSet
from firewall.server.config_helper import FirewallDConfigHelper
from firewall.core.io.functions import check_on_disk_config
from firewall.core.io.icmptype import IcmpType
from firewall.core.io.ipset import IPSet
from firewall.core.io.helper import Helper
from firewall.core.io.direct import Direct
from firewall.dbus_utils import (
    dbus_to_python,
    dbus_introspection_prepare_properties,
    dbus_introspection_add_properties,
    dbus_introspection_add_deprecated,
)
from firewall import errors
from firewall.errors import FirewallError

############################################################################
#
# class FirewallDConfig
#
############################################################################

ConfigPropertiesTuple = collections.namedtuple(
    "ConfigPropertiesTuple",
    [
        "mode",
        "is_deprecated",
        "ignore_set",
    ],
)


CONFIG_PROPERTIES = {
    "DefaultZone": ConfigPropertiesTuple("read", False, False),
    "MinimalMark": ConfigPropertiesTuple("readwrite", True, True),
    "CleanupOnExit": ConfigPropertiesTuple("readwrite", False, False),
    "CleanupModulesOnExit": ConfigPropertiesTuple("readwrite", False, False),
    "IPv6_rpfilter": ConfigPropertiesTuple("readwrite", True, False),
    "IPv6_rpfilter2": ConfigPropertiesTuple("readwrite", False, False),
    "Lockdown": ConfigPropertiesTuple("readwrite", True, True),
    "IndividualCalls": ConfigPropertiesTuple("readwrite", False, False),
    "LogDenied": ConfigPropertiesTuple("readwrite", False, False),
    "AutomaticHelpers": ConfigPropertiesTuple("readwrite", True, True),
    "FirewallBackend": ConfigPropertiesTuple("readwrite", False, False),
    "FlushAllOnReload": ConfigPropertiesTuple("readwrite", False, False),
    "RFC3964_IPv4": ConfigPropertiesTuple("readwrite", False, False),
    "AllowZoneDrifting": ConfigPropertiesTuple("readwrite", True, True),
    "NftablesFlowtable": ConfigPropertiesTuple("readwrite", False, False),
    "NftablesCounters": ConfigPropertiesTuple("readwrite", False, False),
    "NftablesTableOwner": ConfigPropertiesTuple("readwrite", False, False),
    "StrictForwardPorts": ConfigPropertiesTuple("readwrite", False, False),
}


class FirewallDConfig(DbusServiceObject):
    """FirewallD main class"""

    persistent = True
    """ Make FirewallD persistent. """
    default_polkit_auth_required = config.dbus.PK_ACTION_CONFIG
    """ Use config.dbus.PK_ACTION_INFO as a default """

    @handle_exceptions
    def __init__(self, conf, *args, **kwargs):
        super(FirewallDConfig, self).__init__(*args, **kwargs)
        self.config = conf
        self.busname = args[0]
        self.path = args[1]
        self._init_vars()

        dbus_introspection_prepare_properties(
            self,
            config.dbus.DBUS_INTERFACE_CONFIG,
            {prop: d.mode for prop, d in CONFIG_PROPERTIES.items()},
        )

    def _start_watcher(self):
        self.watcher.add_watch_dir(config.FIREWALLD_IPSETS)
        self.watcher.add_watch_dir(config.ETC_FIREWALLD_IPSETS)
        self.watcher.add_watch_dir(config.FIREWALLD_ICMPTYPES)
        self.watcher.add_watch_dir(config.ETC_FIREWALLD_ICMPTYPES)
        self.watcher.add_watch_dir(config.FIREWALLD_HELPERS)
        self.watcher.add_watch_dir(config.ETC_FIREWALLD_HELPERS)
        self.watcher.add_watch_dir(config.FIREWALLD_SERVICES)
        self.watcher.add_watch_dir(config.ETC_FIREWALLD_SERVICES)
        self.watcher.add_watch_dir(config.FIREWALLD_ZONES)
        self.watcher.add_watch_dir(config.ETC_FIREWALLD_ZONES)
        self.watcher.add_watch_dir(config.FIREWALLD_POLICIES)
        self.watcher.add_watch_dir(config.ETC_FIREWALLD_POLICIES)
        # Add watches for combined zone directories
        if os.path.exists(config.ETC_FIREWALLD_ZONES):
            for filename in sorted(os.listdir(config.ETC_FIREWALLD_ZONES)):
                path = "%s/%s" % (config.ETC_FIREWALLD_ZONES, filename)
                if os.path.isdir(path):
                    self.watcher.add_watch_dir(path)
        self.watcher.add_watch_file(config.FIREWALLD_DIRECT)
        self.watcher.add_watch_file(config.FIREWALLD_CONF)

    def _stop_watcher(self):
        self.watcher.clear_timeouts()
        for w in list(self.watcher.get_watches()):
            self.watcher.remove_watch(w)

    def _destroy_watcher(self):
        if not self.watcher:
            return

        self.config.remove_pre_write_hook(self._stop_watcher)
        self.config.remove_post_write_hook(self._start_watcher)
        self._stop_watcher()
        self.watcher = None

    @handle_exceptions
    def _init_vars(self):
        self.ipsets = []
        self.ipset_idx = 0
        self.icmptypes = []
        self.icmptype_idx = 0
        self.services = []
        self.service_idx = 0
        self.zones = []
        self.zone_idx = 0
        self.helpers = []
        self.helper_idx = 0
        self.policy_objects = []
        self.policy_object_idx = 0

        for ipset in self.config.get_ipsets():
            self._addIPSet(self.config.get_ipset(ipset))
        for icmptype in self.config.get_icmptypes():
            self._addIcmpType(self.config.get_icmptype(icmptype))
        for service in self.config.get_services():
            self._addService(self.config.get_service(service))
        for zone in self.config.get_zones():
            self._addZone(self.config.get_zone(zone))
        for helper in self.config.get_helpers():
            self._addHelper(self.config.get_helper(helper))
        for policy in self.config.get_policy_objects():
            self._addPolicy(self.config.get_policy_object(policy))

        self.watcher = Watcher(self.on_disk_config_changed, 5)
        self._start_watcher()
        self.config.add_pre_write_hook(self._stop_watcher)
        self.config.add_post_write_hook(self._start_watcher)

    @handle_exceptions
    def __del__(self):
        pass

    @handle_exceptions
    def reload(self):
        # A reload will trigger a full check config. As such, any pending
        # watcher checks become redundant.
        self._destroy_watcher()

        while len(self.ipsets) > 0:
            item = self.ipsets.pop()
            item.unregister()
            del item
        while len(self.icmptypes) > 0:
            item = self.icmptypes.pop()
            item.unregister()
            del item
        while len(self.services) > 0:
            item = self.services.pop()
            item.unregister()
            del item
        while len(self.zones) > 0:
            item = self.zones.pop()
            item.unregister()
            del item
        while len(self.helpers) > 0:
            item = self.helpers.pop()
            item.unregister()
            del item
        while len(self.policy_objects) > 0:
            item = self.policy_objects.pop()
            item.unregister()
            del item
        self._init_vars()

    def on_disk_config_changed(self, name):
        if not os.path.isfile(name):
            return
        # avoid .bak files, etc.
        (_, ext) = os.path.splitext(name)
        if not ext or ext not in (".xml", ".conf"):
            return

        # log/warn only once
        self._destroy_watcher()

        log.info1(
            "Detected permanent (on-disk) configuration change out-of-band of firewalld!"
        )
        try:
            check_on_disk_config(self.config._fw)
            log.info1("To load the new configuration, reload firewalld:")
            log.info1("  # firewall-cmd --reload")
        except Exception as error:
            log.warning(error)
            log.warning(
                "Unfortunately, the new configuration cannot be loaded by the running firewalld. "
                "This may occur if new configuration was installed that "
                "the running firewalld does not support. Or it may occur "
                "if the configuration is invalid."
            )
            log.warning("It is suggested to check the configuration with command:")
            log.warning("  # firewall-offline-cmd --check-config")
            log.warning("If that succeeds, then please restart the firewalld service:")
            log.warning("  # systemctl restart firewalld")
            log.warning(
                "If the configuration checks fail, then the configuration must be "
                "manually fixed by editing the configuration files."
            )

    @handle_exceptions
    def _addIcmpType(self, obj):
        # TODO: check for idx overflow
        config_icmptype = FirewallDConfigIcmpType(
            self,
            self.config,
            obj,
            self.icmptype_idx,
            self.busname,
            "%s/%d" % (config.dbus.DBUS_PATH_CONFIG_ICMPTYPE, self.icmptype_idx),
        )
        self.icmptypes.append(config_icmptype)
        self.icmptype_idx += 1
        self.IcmpTypeAdded(obj.name)
        return config_icmptype

    @handle_exceptions
    def _updateIcmpType(self, obj):
        for icmptype in self.icmptypes:
            if (
                icmptype.obj.name == obj.name
                and icmptype.obj.path == obj.path
                and icmptype.obj.filename == obj.filename
            ):
                icmptype.obj = obj
                icmptype.Updated(obj.name)

    @handle_exceptions
    def removeIcmpType(self, obj):
        index = 7  # see IMPORT_EXPORT_STRUCTURE in class Zone(IO_Object)
        for zone in self.zones:
            settings = zone.getSettings()
            # if this IcmpType is used in a zone remove it from that zone first
            if obj.name in settings[index]:
                settings[index].remove(obj.name)
                zone.obj = self.config.set_zone_config(zone.obj, settings)
                zone.Updated(zone.obj.name)

        for policy in self.policy_objects:
            settings = policy.getSettings()
            # if this IcmpType is used in a policy remove it from that policy first
            if "icmp_blocks" in settings and obj.name in settings["icmp_blocks"]:
                settings["icmp_blocks"].remove(obj.name)
                policy.obj = self.config.set_policy_object_config_dict(
                    policy.obj, settings
                )
                policy.Updated(policy.obj.name)

        for icmptype in self.icmptypes:
            if icmptype.obj == obj:
                icmptype.Removed(obj.name)
                icmptype.unregister()
                self.icmptypes.remove(icmptype)
                del icmptype

    @handle_exceptions
    def _addService(self, obj):
        # TODO: check for idx overflow
        config_service = FirewallDConfigService(
            self,
            self.config,
            obj,
            self.service_idx,
            self.busname,
            "%s/%d" % (config.dbus.DBUS_PATH_CONFIG_SERVICE, self.service_idx),
        )
        self.services.append(config_service)
        self.service_idx += 1
        self.ServiceAdded(obj.name)
        return config_service

    @handle_exceptions
    def _updateService(self, obj):
        for service in self.services:
            if (
                service.obj.name == obj.name
                and service.obj.path == obj.path
                and service.obj.filename == obj.filename
            ):
                service.obj = obj
                service.Updated(obj.name)

    @handle_exceptions
    def removeService(self, obj):
        index = 5  # see IMPORT_EXPORT_STRUCTURE in class Zone(IO_Object)
        for zone in self.zones:
            settings = zone.getSettings()
            # if this Service is used in a zone remove it from that zone first
            if obj.name in settings[index]:
                settings[index].remove(obj.name)
                zone.obj = self.config.set_zone_config(zone.obj, settings)
                zone.Updated(zone.obj.name)

        for policy in self.policy_objects:
            settings = policy.getSettings()
            # if this Service is used in a policy remove it from that policy first
            if "services" in settings and obj.name in settings["services"]:
                settings["services"].remove(obj.name)
                policy.obj = self.config.set_policy_object_config_dict(
                    policy.obj, settings
                )
                policy.Updated(policy.obj.name)

        for service in self.services:
            if service.obj == obj:
                service.Removed(obj.name)
                service.unregister()
                self.services.remove(service)
                del service

    @handle_exceptions
    def _addZone(self, obj):
        # TODO: check for idx overflow
        config_zone = FirewallDConfigZone(
            self,
            self.config,
            obj,
            self.zone_idx,
            self.busname,
            "%s/%d" % (config.dbus.DBUS_PATH_CONFIG_ZONE, self.zone_idx),
        )
        self.zones.append(config_zone)
        self.zone_idx += 1
        self.ZoneAdded(obj.name)
        return config_zone

    @handle_exceptions
    def _updateZone(self, obj):
        for zone in self.zones:
            if (
                zone.obj.name == obj.name
                and zone.obj.path == obj.path
                and zone.obj.filename == obj.filename
            ):
                zone.obj = obj
                zone.Updated(obj.name)

    @handle_exceptions
    def removeZone(self, obj):
        for zone in self.zones:
            if zone.obj == obj:
                zone.Removed(obj.name)
                zone.unregister()
                self.zones.remove(zone)
                del zone

    @handle_exceptions
    def _addPolicy(self, obj):
        # TODO: check for idx overflow
        config_policy = FirewallDConfigPolicy(
            self,
            self.config,
            obj,
            self.policy_object_idx,
            self.busname,
            "%s/%d" % (config.dbus.DBUS_PATH_CONFIG_POLICY, self.policy_object_idx),
        )
        self.policy_objects.append(config_policy)
        self.policy_object_idx += 1
        self.PolicyAdded(obj.name)
        return config_policy

    @handle_exceptions
    def _updatePolicy(self, obj):
        for policy in self.policy_objects:
            if (
                policy.obj.name == obj.name
                and policy.obj.path == obj.path
                and policy.obj.filename == obj.filename
            ):
                policy.obj = obj
                policy.Updated(obj.name)

    @handle_exceptions
    def removePolicy(self, obj):
        for policy in self.policy_objects:
            if policy.obj == obj:
                policy.Removed(obj.name)
                policy.unregister()
                self.policy_objects.remove(policy)
                del policy

    @handle_exceptions
    def _addIPSet(self, obj):
        # TODO: check for idx overflow
        config_ipset = FirewallDConfigIPSet(
            self,
            self.config,
            obj,
            self.ipset_idx,
            self.busname,
            "%s/%d" % (config.dbus.DBUS_PATH_CONFIG_IPSET, self.ipset_idx),
        )
        self.ipsets.append(config_ipset)
        self.ipset_idx += 1
        self.IPSetAdded(obj.name)
        return config_ipset

    @handle_exceptions
    def _updateIPSet(self, obj):
        for ipset in self.ipsets:
            if (
                ipset.obj.name == obj.name
                and ipset.obj.path == obj.path
                and ipset.obj.filename == obj.filename
            ):
                ipset.obj = obj
                ipset.Updated(obj.name)

    @handle_exceptions
    def removeIPSet(self, obj):
        for ipset in self.ipsets:
            if ipset.obj == obj:
                ipset.Removed(obj.name)
                ipset.unregister()
                self.ipsets.remove(ipset)
                del ipset

    # access check

    @handle_exceptions
    def _addHelper(self, obj):
        # TODO: check for idx overflow
        config_helper = FirewallDConfigHelper(
            self,
            self.config,
            obj,
            self.helper_idx,
            self.busname,
            "%s/%d" % (config.dbus.DBUS_PATH_CONFIG_HELPER, self.helper_idx),
        )
        self.helpers.append(config_helper)
        self.helper_idx += 1
        self.HelperAdded(obj.name)
        return config_helper

    @handle_exceptions
    def _updateHelper(self, obj):
        for helper in self.helpers:
            if (
                helper.obj.name == obj.name
                and helper.obj.path == obj.path
                and helper.obj.filename == obj.filename
            ):
                helper.obj = obj
                helper.Updated(obj.name)

    @handle_exceptions
    def removeHelper(self, obj):
        for helper in self.helpers:
            if helper.obj == obj:
                helper.Removed(obj.name)
                helper.unregister()
                self.helpers.remove(helper)
                del helper

    # access check

    @dbus_handle_exceptions
    def accessCheck(self, sender):
        if self.config._fw._state == "FAILED":
            raise FirewallError(
                errors.RUNNING_BUT_FAILED,
                "Changing permanent configuration is not allowed while "
                "firewalld is in FAILED state. The permanent "
                "configuration must be fixed and then firewalld "
                "restarted. Try `firewall-offline-cmd --check-config`.",
            )

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    # P R O P E R T I E S

    @dbus_handle_exceptions
    def _get_property(self, prop):
        value = self.config.get_firewalld_conf().get(prop)

        if prop == "DefaultZone":
            if value is None:
                value = config.FALLBACK_ZONE
            return dbus.String(value)
        elif prop == "MinimalMark":
            if value is None:
                value = config.FALLBACK_MINIMAL_MARK
            else:
                value = int(value)
            return dbus.Int32(value)
        elif prop == "CleanupOnExit":
            if value is None:
                value = "yes" if config.FALLBACK_CLEANUP_ON_EXIT else "no"
            return dbus.String(value)
        elif prop == "CleanupModulesOnExit":
            if value is None:
                value = "yes" if config.FALLBACK_CLEANUP_MODULES_ON_EXIT else "no"
            return dbus.String(value)
        elif prop == "Lockdown":
            # Deprecated and dropped.
            return dbus.String("no")
        elif prop == "IPv6_rpfilter":
            if value is None or value != "no":
                return dbus.String("yes")
            else:
                return dbus.String("no")
        elif prop == "IPv6_rpfilter2":
            value = self.config.get_firewalld_conf().get("IPv6_rpfilter")
            if value is None:
                value = config.FALLBACK_IPV6_RPFILTER
            return dbus.String(value)
        elif prop == "IndividualCalls":
            if value is None:
                value = "yes" if config.FALLBACK_INDIVIDUAL_CALLS else "no"
            return dbus.String(value)
        elif prop == "LogDenied":
            if value is None:
                value = config.FALLBACK_LOG_DENIED
            return dbus.String(value)
        elif prop == "AutomaticHelpers":
            if value is None:
                value = config.FALLBACK_AUTOMATIC_HELPERS
            return dbus.String(value)
        elif prop == "FirewallBackend":
            if value is None:
                value = config.FALLBACK_FIREWALL_BACKEND
            return dbus.String(value)
        elif prop == "FlushAllOnReload":
            if value is None:
                value = "yes" if config.FALLBACK_FLUSH_ALL_ON_RELOAD else "no"
            return dbus.String(value)
        elif prop == "RFC3964_IPv4":
            if value is None:
                value = "yes" if config.FALLBACK_RFC3964_IPV4 else "no"
            return dbus.String(value)
        elif prop == "AllowZoneDrifting":
            if value is None:
                value = "yes" if config.FALLBACK_ALLOW_ZONE_DRIFTING else "no"
            return dbus.String(value)
        elif prop == "NftablesFlowtable":
            if value is None:
                value = config.FALLBACK_NFTABLES_FLOWTABLE
            return dbus.String(value)
        elif prop == "NftablesCounters":
            if value is None:
                value = "yes" if config.FALLBACK_NFTABLES_COUNTERS else "no"
            return dbus.String(value)
        elif prop == "NftablesTableOwner":
            if value is None:
                value = "yes" if config.FALLBACK_NFTABLES_TABLE_OWNER else "no"
            return dbus.String(value)
        elif prop == "StrictForwardPorts":
            if value is None:
                value = "yes" if config.FALLBACK_STRICT_FORWARD_PORTS else "no"
            return dbus.String(value)
        else:
            raise dbus.exceptions.DBusException(
                "org.freedesktop.DBus.Error.InvalidArgs: "
                "Property '%s' does not exist" % prop
            )

    @dbus_service_method(dbus.PROPERTIES_IFACE, in_signature="ss", out_signature="v")
    @dbus_handle_exceptions
    def Get(self, interface_name, property_name, sender=None):  # pylint: disable=W0613
        # get a property
        interface_name = dbus_to_python(interface_name, str)
        property_name = dbus_to_python(property_name, str)
        log.debug1("config.Get('%s', '%s')", interface_name, property_name)

        if interface_name == config.dbus.DBUS_INTERFACE_CONFIG:
            return self._get_property(property_name)
        elif interface_name in [
            config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
            config.dbus.DBUS_INTERFACE_CONFIG_POLICIES,
        ]:
            raise dbus.exceptions.DBusException(
                "org.freedesktop.DBus.Error.InvalidArgs: "
                "Property '%s' does not exist" % property_name
            )
        else:
            raise dbus.exceptions.DBusException(
                "org.freedesktop.DBus.Error.UnknownInterface: "
                "Interface '%s' does not exist" % interface_name
            )

    @dbus_service_method(dbus.PROPERTIES_IFACE, in_signature="s", out_signature="a{sv}")
    @dbus_handle_exceptions
    def GetAll(self, interface_name, sender=None):  # pylint: disable=W0613
        interface_name = dbus_to_python(interface_name, str)
        log.debug1("config.GetAll('%s')", interface_name)

        ret = {}
        if interface_name == config.dbus.DBUS_INTERFACE_CONFIG:
            for x in CONFIG_PROPERTIES:
                ret[x] = self._get_property(x)
        elif interface_name in [
            config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
            config.dbus.DBUS_INTERFACE_CONFIG_POLICIES,
        ]:
            pass
        else:
            raise dbus.exceptions.DBusException(
                "org.freedesktop.DBus.Error.UnknownInterface: "
                "Interface '%s' does not exist" % interface_name
            )

        return dbus.Dictionary(ret, signature="sv")

    @dbus_polkit_require_auth(config.dbus.PK_ACTION_CONFIG)
    @dbus_service_method(dbus.PROPERTIES_IFACE, in_signature="ssv")
    @dbus_handle_exceptions
    def Set(self, interface_name, property_name, new_value, sender=None):
        interface_name = dbus_to_python(interface_name, str)
        property_name = dbus_to_python(property_name, str)
        new_value = dbus_to_python(new_value)
        log.debug1(
            "config.Set('%s', '%s', '%s')", interface_name, property_name, new_value
        )
        self.accessCheck(sender)

        if interface_name == config.dbus.DBUS_INTERFACE_CONFIG:
            pdata = CONFIG_PROPERTIES.get(property_name)
            if pdata is None or pdata.mode != "readwrite":
                raise dbus.exceptions.DBusException(
                    "org.freedesktop.DBus.Error.InvalidArgs: "
                    "Property '%s' does not exist" % property_name
                )
            if pdata.ignore_set:
                # deprecated fields. Ignore setting them.
                pass
            else:
                if property_name in [
                    "CleanupOnExit",
                    "CleanupModulesOnExit",
                    "IPv6_rpfilter",
                    "IndividualCalls",
                    "FlushAllOnReload",
                    "RFC3964_IPv4",
                    "NftablesCounters",
                    "NftablesTableOwner",
                    "StrictForwardPorts",
                ]:
                    if new_value.lower() not in ["yes", "no", "true", "false"]:
                        raise FirewallError(
                            errors.INVALID_VALUE,
                            "'%s' for %s" % (new_value, property_name),
                        )
                    config_name = property_name
                elif property_name == "LogDenied":
                    if new_value not in config.LOG_DENIED_VALUES:
                        raise FirewallError(
                            errors.INVALID_VALUE,
                            "'%s' for %s" % (new_value, property_name),
                        )
                    config_name = property_name
                elif property_name == "FirewallBackend":
                    if new_value not in config.FIREWALL_BACKEND_VALUES:
                        raise FirewallError(
                            errors.INVALID_VALUE,
                            "'%s' for %s" % (new_value, property_name),
                        )
                    config_name = property_name
                elif property_name == "IPv6_rpfilter2":
                    if new_value not in config.IPV6_RPFILTER_VALUES:
                        raise FirewallError(
                            errors.INVALID_VALUE,
                            "'%s' for %s" % (new_value, property_name),
                        )
                    config_name = "IPv6_rpfilter"
                else:
                    raise errors.BugError(f'Unhandled property_name "{property_name}"')

                self.config.get_firewalld_conf().set(config_name, new_value)
                self.config.call_pre_write_hooks()
                self.config.get_firewalld_conf().write()
                self.config.call_post_write_hooks()
                self.PropertiesChanged(interface_name, {property_name: new_value}, [])
        elif interface_name in [
            config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
            config.dbus.DBUS_INTERFACE_CONFIG_POLICIES,
        ]:
            raise dbus.exceptions.DBusException(
                "org.freedesktop.DBus.Error.InvalidArgs: "
                "Property '%s' does not exist" % property_name
            )
        else:
            raise dbus.exceptions.DBusException(
                "org.freedesktop.DBus.Error.UnknownInterface: "
                "Interface '%s' does not exist" % interface_name
            )

    @dbus_service_signal(dbus.PROPERTIES_IFACE, signature="sa{sv}as")
    def PropertiesChanged(
        self, interface_name, changed_properties, invalidated_properties
    ):
        interface_name = dbus_to_python(interface_name, str)
        changed_properties = dbus_to_python(changed_properties)
        invalidated_properties = dbus_to_python(invalidated_properties)
        log.debug1(
            "config.PropertiesChanged('%s', '%s', '%s')",
            interface_name,
            changed_properties,
            invalidated_properties,
        )

    @dbus_polkit_require_auth(config.dbus.PK_ACTION_INFO)
    @dbus_service_method(dbus.INTROSPECTABLE_IFACE, out_signature="s")
    @dbus_handle_exceptions
    def Introspect(self, sender=None):  # pylint: disable=W0613
        log.debug2("config.Introspect()")

        data = super(FirewallDConfig, self).Introspect(
            self.path, self.busname.get_bus()
        )
        data = dbus_introspection_add_properties(
            self, data, config.dbus.DBUS_INTERFACE_CONFIG
        )

        for interface in [
            config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
            config.dbus.DBUS_INTERFACE_CONFIG_POLICIES,
        ]:
            data = dbus_introspection_add_deprecated(
                self,
                data,
                interface,
                dbus_service_method_deprecated.deprecated,
                dbus_service_signal_deprecated.deprecated,
            )

        return data

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    # policies

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_POLICIES,
        out_signature="(asasasai)",
        is_deprecated=True,
    )
    def getLockdownWhitelist(self, sender=None):  # pylint: disable=W0613
        return ([], [], [], [0])

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_POLICIES,
        in_signature="(asasasai)",
        is_deprecated=True,
    )
    def setLockdownWhitelist(self, settings, sender=None):  # pylint: disable=W0613
        pass

    @dbus_service_signal(
        config.dbus.DBUS_INTERFACE_CONFIG_POLICIES,
        is_deprecated=True,
    )
    def LockdownWhitelistUpdated(self):
        pass

    # command

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_POLICIES,
        in_signature="s",
        is_deprecated=True,
    )
    def addLockdownWhitelistCommand(self, command, sender=None):
        pass

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_POLICIES,
        in_signature="s",
        is_deprecated=True,
    )
    def removeLockdownWhitelistCommand(self, command, sender=None):
        pass

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_POLICIES,
        in_signature="s",
        out_signature="b",
        is_deprecated=True,
    )
    def queryLockdownWhitelistCommand(self, command, sender=None):
        return False

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_POLICIES,
        out_signature="as",
        is_deprecated=True,
    )
    def getLockdownWhitelistCommands(self, sender=None):  # pylint: disable=W0613
        return []

    # context

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_POLICIES,
        in_signature="s",
        is_deprecated=True,
    )
    def addLockdownWhitelistContext(self, context, sender=None):
        pass

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_POLICIES,
        in_signature="s",
        is_deprecated=True,
    )
    def removeLockdownWhitelistContext(self, context, sender=None):
        pass

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_POLICIES,
        in_signature="s",
        out_signature="b",
        is_deprecated=True,
    )
    def queryLockdownWhitelistContext(
        self, context, sender=None
    ):  # pylint: disable=W0613
        return False

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_POLICIES,
        out_signature="as",
        is_deprecated=True,
    )
    def getLockdownWhitelistContexts(self, sender=None):  # pylint: disable=W0613
        return []

    # user

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_POLICIES,
        in_signature="s",
        is_deprecated=True,
    )
    def addLockdownWhitelistUser(self, user, sender=None):
        pass

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_POLICIES,
        in_signature="s",
        is_deprecated=True,
    )
    def removeLockdownWhitelistUser(self, user, sender=None):
        pass

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_POLICIES,
        in_signature="s",
        out_signature="b",
        is_deprecated=True,
    )
    def queryLockdownWhitelistUser(self, user, sender=None):  # pylint: disable=W0613
        return False

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_POLICIES,
        out_signature="as",
        is_deprecated=True,
    )
    def getLockdownWhitelistUsers(self, sender=None):  # pylint: disable=W0613
        return []

    # uid

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_POLICIES,
        in_signature="i",
        is_deprecated=True,
    )
    def addLockdownWhitelistUid(self, uid, sender=None):
        pass

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_POLICIES,
        in_signature="i",
        is_deprecated=True,
    )
    def removeLockdownWhitelistUid(self, uid, sender=None):
        pass

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_POLICIES,
        in_signature="i",
        out_signature="b",
        is_deprecated=True,
    )
    def queryLockdownWhitelistUid(self, uid, sender=None):  # pylint: disable=W0613
        return False

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_POLICIES,
        out_signature="ai",
        is_deprecated=True,
    )
    def getLockdownWhitelistUids(self, sender=None):  # pylint: disable=W0613
        return []

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    # I P S E T S

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG, out_signature="ao")
    @dbus_handle_exceptions
    def listIPSets(self, sender=None):  # pylint: disable=W0613
        """list ipsets objects paths"""
        log.debug1("config.listIPSets()")
        return self.ipsets

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG, out_signature="as")
    @dbus_handle_exceptions
    def getIPSetNames(self, sender=None):  # pylint: disable=W0613
        """get ipset names"""
        log.debug1("config.getIPSetNames()")
        ipsets = []
        for obj in self.ipsets:
            ipsets.append(obj.obj.name)
        return sorted(ipsets)

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG, in_signature="s", out_signature="o"
    )
    @dbus_handle_exceptions
    def getIPSetByName(self, ipset, sender=None):  # pylint: disable=W0613
        """object path of ipset with given name"""
        ipset = dbus_to_python(ipset, str)
        log.debug1("config.getIPSetByName('%s')", ipset)
        for obj in self.ipsets:
            if obj.obj.name == ipset:
                return obj
        raise FirewallError(errors.INVALID_IPSET, ipset)

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG,
        in_signature="s" + IPSet.DBUS_SIGNATURE,
        out_signature="o",
    )
    @dbus_handle_exceptions
    def addIPSet(self, ipset, settings, sender=None):
        """add ipset with given name and settings"""
        ipset = dbus_to_python(ipset, str)
        settings = dbus_to_python(settings)
        log.debug1("config.addIPSet('%s')", ipset)
        self.accessCheck(sender)
        obj = self.config.new_ipset(ipset, settings)
        config_ipset = self._addIPSet(obj)
        return config_ipset

    @dbus_service_signal(config.dbus.DBUS_INTERFACE_CONFIG, signature="s")
    @dbus_handle_exceptions
    def IPSetAdded(self, ipset):
        ipset = dbus_to_python(ipset, str)
        log.debug1("config.IPSetAdded('%s')" % (ipset))

    # I C M P T Y P E S

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG, out_signature="ao")
    @dbus_handle_exceptions
    def listIcmpTypes(self, sender=None):  # pylint: disable=W0613
        """list icmptypes objects paths"""
        log.debug1("config.listIcmpTypes()")
        return self.icmptypes

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG, out_signature="as")
    @dbus_handle_exceptions
    def getIcmpTypeNames(self, sender=None):  # pylint: disable=W0613
        """get icmptype names"""
        log.debug1("config.getIcmpTypeNames()")
        icmptypes = []
        for obj in self.icmptypes:
            icmptypes.append(obj.obj.name)
        return sorted(icmptypes)

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG, in_signature="s", out_signature="o"
    )
    @dbus_handle_exceptions
    def getIcmpTypeByName(self, icmptype, sender=None):  # pylint: disable=W0613
        """object path of icmptype with given name"""
        icmptype = dbus_to_python(icmptype, str)
        log.debug1("config.getIcmpTypeByName('%s')", icmptype)
        for obj in self.icmptypes:
            if obj.obj.name == icmptype:
                return obj
        raise FirewallError(errors.INVALID_ICMPTYPE, icmptype)

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG,
        in_signature="s" + IcmpType.DBUS_SIGNATURE,
        out_signature="o",
    )
    @dbus_handle_exceptions
    def addIcmpType(self, icmptype, settings, sender=None):
        """add icmptype with given name and settings"""
        icmptype = dbus_to_python(icmptype, str)
        settings = dbus_to_python(settings)
        log.debug1("config.addIcmpType('%s')", icmptype)
        self.accessCheck(sender)
        obj = self.config.new_icmptype(icmptype, settings)
        config_icmptype = self._addIcmpType(obj)
        return config_icmptype

    @dbus_service_signal(config.dbus.DBUS_INTERFACE_CONFIG, signature="s")
    @dbus_handle_exceptions
    def IcmpTypeAdded(self, icmptype):
        log.debug1("config.IcmpTypeAdded('%s')" % (icmptype))

    # S E R V I C E S

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG, out_signature="ao")
    @dbus_handle_exceptions
    def listServices(self, sender=None):  # pylint: disable=W0613
        """list services objects paths"""
        log.debug1("config.listServices()")
        return self.services

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG, out_signature="as")
    @dbus_handle_exceptions
    def getServiceNames(self, sender=None):  # pylint: disable=W0613
        """get service names"""
        log.debug1("config.getServiceNames()")
        services = []
        for obj in self.services:
            services.append(obj.obj.name)
        return sorted(services)

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG, in_signature="s", out_signature="o"
    )
    @dbus_handle_exceptions
    def getServiceByName(self, service, sender=None):  # pylint: disable=W0613
        """object path of service with given name"""
        service = dbus_to_python(service, str)
        log.debug1("config.getServiceByName('%s')", service)
        for obj in self.services:
            if obj.obj.name == service:
                return obj
        raise FirewallError(errors.INVALID_SERVICE, service)

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG,
        in_signature="s(sssa(ss)asa{ss}asa(ss))",
        out_signature="o",
    )
    @dbus_handle_exceptions
    def addService(self, service, settings, sender=None):
        """add service with given name and settings"""
        service = dbus_to_python(service, str)
        settings = dbus_to_python(settings)
        log.debug1("config.addService('%s')", service)
        self.accessCheck(sender)
        obj = self.config.new_service(service, settings)
        config_service = self._addService(obj)
        return config_service

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG, in_signature="sa{sv}", out_signature="o"
    )
    @dbus_handle_exceptions
    def addService2(self, service, settings, sender=None):
        """add service with given name and settings"""
        service = dbus_to_python(service, str)
        settings = dbus_to_python(settings)
        log.debug1("config.addService2('%s')", service)
        self.accessCheck(sender)
        obj = self.config.new_service_dict(service, settings)
        config_service = self._addService(obj)
        return config_service

    @dbus_service_signal(config.dbus.DBUS_INTERFACE_CONFIG, signature="s")
    @dbus_handle_exceptions
    def ServiceAdded(self, service):
        log.debug1("config.ServiceAdded('%s')" % (service))

    # Z O N E S

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG, out_signature="ao")
    @dbus_handle_exceptions
    def listZones(self, sender=None):  # pylint: disable=W0613
        """list zones objects paths"""
        log.debug1("config.listZones()")
        return self.zones

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG, out_signature="as")
    @dbus_handle_exceptions
    def getZoneNames(self, sender=None):  # pylint: disable=W0613
        """get zone names"""
        log.debug1("config.getZoneNames()")
        zones = []
        for obj in self.zones:
            zones.append(obj.obj.name)
        return sorted(zones)

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG, in_signature="s", out_signature="o"
    )
    @dbus_handle_exceptions
    def getZoneByName(self, zone, sender=None):  # pylint: disable=W0613
        """object path of zone with given name"""
        zone = dbus_to_python(zone, str)
        log.debug1("config.getZoneByName('%s')", zone)
        for obj in self.zones:
            if obj.obj.name == zone:
                return obj
        raise FirewallError(errors.INVALID_ZONE, zone)

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG, in_signature="s", out_signature="s"
    )
    @dbus_handle_exceptions
    def getZoneOfInterface(self, iface, sender=None):  # pylint: disable=W0613
        """name of zone the given interface belongs to"""
        iface = dbus_to_python(iface, str)
        log.debug1("config.getZoneOfInterface('%s')", iface)
        ret = []
        for obj in self.zones:
            if iface in obj.obj.interfaces:
                ret.append(obj.obj.name)
        if len(ret) > 1:
            # Even it shouldn't happen, it's actually possible that
            # the same interface is in several zone XML files
            return " ".join(
                ret
            ) + "  (ERROR: interface '%s' is in %s zone XML files, can be only in one)" % (
                iface,
                len(ret),
            )
        return ret[0] if ret else ""

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG, in_signature="s", out_signature="s"
    )
    @dbus_handle_exceptions
    def getZoneOfSource(self, source, sender=None):  # pylint: disable=W0613
        """name of zone the given source belongs to"""
        source = dbus_to_python(source, str)
        log.debug1("config.getZoneOfSource('%s')", source)
        ret = []
        for obj in self.zones:
            if source in obj.obj.sources:
                ret.append(obj.obj.name)
        if len(ret) > 1:
            # Even it shouldn't happen, it's actually possible that
            # the same source is in several zone XML files
            return " ".join(
                ret
            ) + "  (ERROR: source '%s' is in %s zone XML files, can be only in one)" % (
                source,
                len(ret),
            )
        return ret[0] if ret else ""

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG,
        in_signature="s(sssbsasa(ss)asba(ssss)asasasasa(ss)b)",
        out_signature="o",
    )
    @dbus_handle_exceptions
    def addZone(self, zone, settings, sender=None):
        """add zone with given name and settings"""
        zone = dbus_to_python(zone, str)
        settings = dbus_to_python(settings)
        log.debug1("config.addZone('%s')", zone)
        self.accessCheck(sender)
        if settings[4] == "default":
            # convert to list, fix target, convert back to tuple
            _settings = list(settings)
            _settings[4] = DEFAULT_ZONE_TARGET
            settings = tuple(_settings)
        obj = self.config.new_zone(zone, settings)
        config_zone = self._addZone(obj)
        return config_zone

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG, in_signature="sa{sv}", out_signature="o"
    )
    @dbus_handle_exceptions
    def addZone2(self, zone, settings, sender=None):
        """add zone with given name and settings"""
        zone = dbus_to_python(zone, str)
        settings = dbus_to_python(settings)
        log.debug1("config.addZone('%s')", zone)
        self.accessCheck(sender)
        if "target" in settings and settings["target"] == "default":
            settings["target"] = DEFAULT_ZONE_TARGET
        obj = self.config.new_zone_dict(zone, settings)
        config_zone = self._addZone(obj)
        return config_zone

    @dbus_service_signal(config.dbus.DBUS_INTERFACE_CONFIG, signature="s")
    @dbus_handle_exceptions
    def ZoneAdded(self, zone):
        log.debug1("config.ZoneAdded('%s')" % (zone))

    # policies

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG, out_signature="ao")
    @dbus_handle_exceptions
    def listPolicies(self, sender=None):
        """list policies objects paths"""
        log.debug1("config.listPolicies()")
        return self.policy_objects

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG, out_signature="as")
    @dbus_handle_exceptions
    def getPolicyNames(self, sender=None):
        """get policy names"""
        log.debug1("config.getPolicyNames()")
        policies = []
        for obj in self.policy_objects:
            policies.append(obj.obj.name)
        return sorted(policies)

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG, in_signature="s", out_signature="o"
    )
    @dbus_handle_exceptions
    def getPolicyByName(self, policy, sender=None):
        """object path of policy with given name"""
        policy = dbus_to_python(policy, str)
        log.debug1("config.getPolicyByName('%s')", policy)
        for obj in self.policy_objects:
            if obj.obj.name == policy:
                return obj
        raise FirewallError(errors.INVALID_POLICY, policy)

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG, in_signature="sa{sv}", out_signature="o"
    )
    @dbus_handle_exceptions
    def addPolicy(self, policy, settings, sender=None):
        """add policy with given name and settings"""
        policy = dbus_to_python(policy, str)
        settings = dbus_to_python(settings)
        log.debug1("config.addPolicy('%s')", policy)
        self.accessCheck(sender)
        obj = self.config.new_policy_object_dict(policy, settings)
        config_policy = self._addPolicy(obj)
        return config_policy

    @dbus_service_signal(config.dbus.DBUS_INTERFACE_CONFIG, signature="s")
    @dbus_handle_exceptions
    def PolicyAdded(self, policy):
        log.debug1("config.PolicyAdded('%s')" % (policy))

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    # H E L P E R S

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG, out_signature="ao")
    @dbus_handle_exceptions
    def listHelpers(self, sender=None):  # pylint: disable=W0613
        """list helpers objects paths"""
        log.debug1("config.listHelpers()")
        return self.helpers

    @dbus_service_method(config.dbus.DBUS_INTERFACE_CONFIG, out_signature="as")
    @dbus_handle_exceptions
    def getHelperNames(self, sender=None):  # pylint: disable=W0613
        """get helper names"""
        log.debug1("config.getHelperNames()")
        helpers = []
        for obj in self.helpers:
            helpers.append(obj.obj.name)
        return sorted(helpers)

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG, in_signature="s", out_signature="o"
    )
    @dbus_handle_exceptions
    def getHelperByName(self, helper, sender=None):  # pylint: disable=W0613
        """object path of helper with given name"""
        helper = dbus_to_python(helper, str)
        log.debug1("config.getHelperByName('%s')", helper)
        for obj in self.helpers:
            if obj.obj.name == helper:
                return obj
        raise FirewallError(errors.INVALID_HELPER, helper)

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG,
        in_signature="s" + Helper.DBUS_SIGNATURE,
        out_signature="o",
    )
    @dbus_handle_exceptions
    def addHelper(self, helper, settings, sender=None):
        """add helper with given name and settings"""
        helper = dbus_to_python(helper, str)
        settings = dbus_to_python(settings)
        log.debug1("config.addHelper('%s')", helper)
        self.accessCheck(sender)
        obj = self.config.new_helper(helper, settings)
        config_helper = self._addHelper(obj)
        return config_helper

    @dbus_service_signal(config.dbus.DBUS_INTERFACE_CONFIG, signature="s")
    @dbus_handle_exceptions
    def HelperAdded(self, helper):
        helper = dbus_to_python(helper, str)
        log.debug1("config.HelperAdded('%s')" % (helper))

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    # DIRECT

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
        out_signature=Direct.DBUS_SIGNATURE,
        is_deprecated=True,
    )
    @dbus_handle_exceptions
    def getSettings(self, sender=None):  # pylint: disable=W0613
        # returns list ipv, table, list of chains
        log.debug1("config.direct.getSettings()")
        return self.config.get_direct().export_config()

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
        in_signature=Direct.DBUS_SIGNATURE,
        is_deprecated=True,
    )
    @dbus_handle_exceptions
    def update(self, settings, sender=None):  # pylint: disable=W0613
        # returns list ipv, table, list of chains
        log.debug1("config.direct.update()")
        settings = dbus_to_python(settings)
        self.config.get_direct().import_config(settings, {})
        self.config.call_pre_write_hooks()
        self.config.get_direct().write()
        self.config.call_post_write_hooks()
        self.Updated()

    @dbus_service_signal(
        config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
        is_deprecated=True,
    )
    @dbus_handle_exceptions
    def Updated(self):
        log.debug1("config.direct.Updated()")

    # chain

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
        in_signature="sss",
        is_deprecated=True,
    )
    @dbus_handle_exceptions
    def addChain(self, ipv, table, chain, sender=None):
        ipv = dbus_to_python(ipv)
        table = dbus_to_python(table)
        chain = dbus_to_python(chain)
        log.debug1("config.direct.addChain('%s', '%s', '%s')" % (ipv, table, chain))
        self.accessCheck(sender)
        idx = tuple((ipv, table, chain))
        settings = list(self.getSettings())
        if idx in settings[0]:
            raise FirewallError(
                errors.ALREADY_ENABLED,
                "chain '%s' already is in '%s:%s'" % (chain, ipv, table),
            )
        settings[0].append(idx)
        self.update(settings)

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
        in_signature="sss",
        is_deprecated=True,
    )
    @dbus_handle_exceptions
    def removeChain(self, ipv, table, chain, sender=None):
        ipv = dbus_to_python(ipv)
        table = dbus_to_python(table)
        chain = dbus_to_python(chain)
        log.debug1("config.direct.removeChain('%s', '%s', '%s')" % (ipv, table, chain))
        self.accessCheck(sender)
        idx = tuple((ipv, table, chain))
        settings = list(self.getSettings())
        if idx not in settings[0]:
            raise FirewallError(
                errors.NOT_ENABLED, "chain '%s' is not in '%s:%s'" % (chain, ipv, table)
            )
        settings[0].remove(idx)
        self.update(settings)

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
        in_signature="sss",
        out_signature="b",
        is_deprecated=True,
    )
    @dbus_handle_exceptions
    def queryChain(self, ipv, table, chain, sender=None):  # pylint: disable=W0613
        ipv = dbus_to_python(ipv)
        table = dbus_to_python(table)
        chain = dbus_to_python(chain)
        log.debug1("config.direct.queryChain('%s', '%s', '%s')" % (ipv, table, chain))
        idx = tuple((ipv, table, chain))
        return idx in self.getSettings()[0]

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
        in_signature="ss",
        out_signature="as",
        is_deprecated=True,
    )
    @dbus_handle_exceptions
    def getChains(self, ipv, table, sender=None):  # pylint: disable=W0613
        ipv = dbus_to_python(ipv)
        table = dbus_to_python(table)
        log.debug1("config.direct.getChains('%s', '%s')" % (ipv, table))
        ret = []
        for idx in self.getSettings()[0]:
            if idx[0] == ipv and idx[1] == table:
                ret.append(idx[2])
        return ret

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
        in_signature="",
        out_signature="a(sss)",
        is_deprecated=True,
    )
    @dbus_handle_exceptions
    def getAllChains(self, sender=None):  # pylint: disable=W0613
        log.debug1("config.direct.getAllChains()")
        return self.getSettings()[0]

    # rule

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
        in_signature="sssias",
        is_deprecated=True,
    )
    @dbus_handle_exceptions
    def addRule(
        self, ipv, table, chain, priority, args, sender=None
    ):  # pylint: disable=R0913
        ipv = dbus_to_python(ipv)
        table = dbus_to_python(table)
        chain = dbus_to_python(chain)
        priority = dbus_to_python(priority)
        args = dbus_to_python(args)
        log.debug1(
            "config.direct.addRule('%s', '%s', '%s', %d, '%s')"
            % (ipv, table, chain, priority, "','".join(args))
        )
        self.accessCheck(sender)
        idx = (ipv, table, chain, priority, args)
        settings = list(self.getSettings())
        if idx in settings[1]:
            raise FirewallError(
                errors.ALREADY_ENABLED,
                "rule '%s' already is in '%s:%s:%s'" % (args, ipv, table, chain),
            )
        settings[1].append(idx)
        self.update(tuple(settings))

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
        in_signature="sssias",
        is_deprecated=True,
    )
    @dbus_handle_exceptions
    def removeRule(
        self, ipv, table, chain, priority, args, sender=None
    ):  # pylint: disable=R0913
        ipv = dbus_to_python(ipv)
        table = dbus_to_python(table)
        chain = dbus_to_python(chain)
        priority = dbus_to_python(priority)
        args = dbus_to_python(args)
        log.debug1(
            "config.direct.removeRule('%s', '%s', '%s', %d, '%s')"
            % (ipv, table, chain, priority, "','".join(args))
        )
        self.accessCheck(sender)
        idx = (ipv, table, chain, priority, args)
        settings = list(self.getSettings())
        if idx not in settings[1]:
            raise FirewallError(
                errors.NOT_ENABLED,
                "rule '%s' is not in '%s:%s:%s'" % (args, ipv, table, chain),
            )
        settings[1].remove(idx)
        self.update(tuple(settings))

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
        in_signature="sssias",
        out_signature="b",
        is_deprecated=True,
    )
    @dbus_handle_exceptions
    def queryRule(
        self, ipv, table, chain, priority, args, sender=None
    ):  # pylint: disable=W0613,R0913
        ipv = dbus_to_python(ipv)
        table = dbus_to_python(table)
        chain = dbus_to_python(chain)
        priority = dbus_to_python(priority)
        args = dbus_to_python(args)
        log.debug1(
            "config.direct.queryRule('%s', '%s', '%s', %d, '%s')"
            % (ipv, table, chain, priority, "','".join(args))
        )
        idx = (ipv, table, chain, priority, args)
        return idx in self.getSettings()[1]

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
        in_signature="sss",
        is_deprecated=True,
    )
    @dbus_handle_exceptions
    def removeRules(self, ipv, table, chain, sender=None):
        ipv = dbus_to_python(ipv)
        table = dbus_to_python(table)
        chain = dbus_to_python(chain)
        log.debug1(
            "config.direct.removeRules('%s', '%s', '%s')"
            % (
                ipv,
                table,
                chain,
            )
        )
        self.accessCheck(sender)
        settings = list(self.getSettings())
        for rule in settings[1][:]:
            if (ipv, table, chain) == (rule[0], rule[1], rule[2]):
                settings[1].remove(rule)
        self.update(tuple(settings))

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
        in_signature="sss",
        out_signature="a(ias)",
        is_deprecated=True,
    )
    @dbus_handle_exceptions
    def getRules(self, ipv, table, chain, sender=None):  # pylint: disable=W0613
        ipv = dbus_to_python(ipv)
        table = dbus_to_python(table)
        chain = dbus_to_python(chain)
        log.debug1("config.direct.getRules('%s', '%s', '%s')" % (ipv, table, chain))
        ret = []
        for idx in self.getSettings()[1]:
            if idx[0] == ipv and idx[1] == table and idx[2] == chain:
                ret.append((idx[3], idx[4]))
        return ret

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
        in_signature="",
        out_signature="a(sssias)",
        is_deprecated=True,
    )
    @dbus_handle_exceptions
    def getAllRules(self, sender=None):  # pylint: disable=W0613
        log.debug1("config.direct.getAllRules()")
        return self.getSettings()[1]

    # passthrough

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
        in_signature="sas",
        is_deprecated=True,
    )
    @dbus_handle_exceptions
    def addPassthrough(self, ipv, args, sender=None):
        ipv = dbus_to_python(ipv)
        args = dbus_to_python(args)
        log.debug1("config.direct.addPassthrough('%s', '%s')" % (ipv, "','".join(args)))
        self.accessCheck(sender)
        idx = (ipv, args)
        settings = list(self.getSettings())
        if idx in settings[2]:
            raise FirewallError(
                errors.ALREADY_ENABLED, "passthrough '%s', '%s'" % (ipv, args)
            )
        settings[2].append(idx)
        self.update(settings)

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
        in_signature="sas",
        is_deprecated=True,
    )
    @dbus_handle_exceptions
    def removePassthrough(self, ipv, args, sender=None):
        ipv = dbus_to_python(ipv)
        args = dbus_to_python(args)
        log.debug1(
            "config.direct.removePassthrough('%s', '%s')" % (ipv, "','".join(args))
        )
        self.accessCheck(sender)
        idx = (ipv, args)
        settings = list(self.getSettings())
        if idx not in settings[2]:
            raise FirewallError(
                errors.NOT_ENABLED, "passthrough '%s', '%s'" % (ipv, args)
            )
        settings[2].remove(idx)
        self.update(settings)

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
        in_signature="sas",
        out_signature="b",
        is_deprecated=True,
    )
    @dbus_handle_exceptions
    def queryPassthrough(self, ipv, args, sender=None):  # pylint: disable=W0613
        ipv = dbus_to_python(ipv)
        args = dbus_to_python(args)
        log.debug1(
            "config.direct.queryPassthrough('%s', '%s')" % (ipv, "','".join(args))
        )
        idx = (ipv, args)
        return idx in self.getSettings()[2]

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
        in_signature="s",
        out_signature="aas",
        is_deprecated=True,
    )
    @dbus_handle_exceptions
    def getPassthroughs(self, ipv, sender=None):  # pylint: disable=W0613
        ipv = dbus_to_python(ipv)
        log.debug1("config.direct.getPassthroughs('%s')" % (ipv))
        ret = []
        for idx in self.getSettings()[2]:
            if idx[0] == ipv:
                ret.append(idx[1])
        return ret

    @dbus_service_method(
        config.dbus.DBUS_INTERFACE_CONFIG_DIRECT,
        out_signature="a(sas)",
        is_deprecated=True,
    )
    @dbus_handle_exceptions
    def getAllPassthroughs(self, sender=None):  # pylint: disable=W0613
        log.debug1("config.direct.getAllPassthroughs()")
        return self.getSettings()[2]



#
# SPDX-License-Identifier: GPL-2.0-or-later

from gi.repository import GLib

import copy
import dataclasses

from firewall.core.logger import log
from firewall.functions import (
    portStr,
    checkIPnMask,
    checkIP6nMask,
    checkProtocol,
    enable_ip_forwarding,
    check_single_address,
    portInPortRange,
    get_nf_conntrack_short_name,
    coalescePortRange,
    breakPortRange,
    checkTcpMssClamp,
)
from firewall.core.rich import (
    Rich_Rule,
    Rich_Accept,
    Rich_Service,
    Rich_Port,
    Rich_Protocol,
    Rich_Masquerade,
    Rich_ForwardPort,
    Rich_SourcePort,
    Rich_IcmpBlock,
    Rich_IcmpType,
    Rich_Tcp_Mss_Clamp,
)
from firewall import errors
from firewall.errors import FirewallError
from firewall.core.base import SOURCE_IPSET_TYPES


class FirewallPolicy:
    def __init__(self, fw):
        self._fw = fw
        self._chains = {}
        self._policies = {}
        self._timeouts = {}

    def __repr__(self):
        return "%s(%r, %r)" % (self.__class__, self._chains, self._policies)

    def cleanup(self):
        self._chains.clear()
        self._policies.clear()
        for _id in self._timeouts:
            self.removeTimeout(_id)

    # transaction

    def with_transaction(self, *args, **kwargs):
        ctx = self._fw.with_transaction(*args, **kwargs)
        if ctx.is_temporary:
            ctx.transaction.add_pre(self._fw.full_check_config)
        return ctx

    # policies

    def get_policies(self):
        return sorted(self._policies.keys())

    def get_policies_not_derived_from_zone(self):
        policies = []
        for p in self.get_policies():
            p_obj = self.get_policy(p)
            if not p_obj.derived_from_zone:
                policies.append(p)
        return sorted(policies)

    def get_active_policies_not_derived_from_zone(self):
        active_policies = []
        for policy in self.get_policies_not_derived_from_zone():
            p_obj = self.get_policy(policy)
            if p_obj.disable:
                continue
            if (
                set(p_obj.ingress_zones)
                & set(self._fw.zone.get_active_zones() + ["HOST", "ANY"])
            ) and (
                set(p_obj.egress_zones)
                & set(self._fw.zone.get_active_zones() + ["HOST", "ANY"])
            ):
                active_policies.append(policy)

        return active_policies

    def get_applied_policies(self):
        policies = self.get_active_policies_not_derived_from_zone()
        return list(filter(lambda x: self.get_policy(x).applied, policies))

    def get_unapplied_policies(self):
        policies = self.get_active_policies_not_derived_from_zone()
        return list(filter(lambda x: not self.get_policy(x).applied, policies))

    def get_policy(self, policy):
        p = self._fw.check_policy(policy)
        return self._policies[p]

    def add_policy(self, obj):
        self._policies[obj.name] = obj

    def remove_policy(self, policy):
        obj = self._policies[policy]
        if obj.applied:
            with self.with_transaction() as transaction:
                self.unapply_policy_settings(policy, transaction)
        del self._policies[policy]

    def apply_policies(self, transaction):
        for policy in self.get_policies():
            p_obj = self._policies[policy]
            if p_obj.derived_from_zone:
                continue
            if policy in self.get_active_policies_not_derived_from_zone():
                log.debug1("Applying policy '%s'", policy)
                self.apply_policy_settings(policy, transaction)

    def set_policy_applied(self, policy, applied):
        obj = self._policies[policy]
        obj.applied = applied

    def _policy_settings(self, enable, policy, transaction):
        _policy = self._fw.check_policy(policy)
        obj = self._policies[_policy]
        if (enable and obj.applied) or (not enable and not obj.applied):
            return
        if enable:
            obj.applied = True

        if enable:
            # build the base chain layout of the policy
            for table, chain in self._get_table_chains_for_policy_dispatch(policy):
                self.gen_chain_rules(policy, True, table, chain, transaction)

        for key in [
            "services",
            "ports",
            "masquerade",
            "forward_ports",
            "source_ports",
            "icmp_blocks",
            "rules",
            "protocols",
            "icmp_block_inversion",
            "ingress_zones",
            "egress_zones",
        ]:
            args_list = getattr(self.get_policy(policy), key)
            if isinstance(args_list, bool):
                if not ((enable and args_list) or (not enable and args_list)):
                    continue
                args_list = [args_list]
            if key == "rules":
                args_list = sorted(args_list)
            for args in args_list:
                if key == "icmp_blocks":
                    self._icmp_block(enable, _policy, args, transaction)
                elif key == "icmp_block_inversion":
                    continue
                elif key == "forward_ports":
                    self._forward_port(enable, _policy, transaction, *args)
                elif key == "services":
                    self._service(enable, _policy, args, transaction)
                elif key == "ports":
                    self._port(enable, _policy, args[0], args[1], transaction)
                elif key == "protocols":
                    self._protocol(enable, _policy, args, transaction)
                elif key == "source_ports":
                    self._source_port(enable, _policy, args[0], args[1], transaction)
                elif key == "masquerade":
                    self._masquerade(enable, _policy, transaction)
                elif key == "rules":
                    self.__rule(enable, _policy, args, transaction)
                elif key == "ingress_zones":
                    self._ingress_zone(enable, _policy, args, transaction)
                elif key == "egress_zones":
                    # key off ingress zones, which also considers egress zones
                    continue
                else:
                    log.warning(
                        "Policy '%s': Unknown setting '%s:%s', " "unable to apply",
                        policy,
                        key,
                        args,
                    )

        if not enable:
            for table, chain in self._get_table_chains_for_policy_dispatch(policy):
                self.gen_chain_rules(policy, False, table, chain, transaction)
            obj.applied = False

    def apply_policy_settings(self, policy, transaction):
        self._policy_settings(True, policy, transaction)

    def try_apply_policy_settings(self, policy, use_transaction=None):
        if policy in self.get_active_policies_not_derived_from_zone():
            with self.with_transaction(use_transaction) as transaction:
                self.apply_policy_settings(policy, transaction)

    def unapply_policy_settings(self, policy, transaction):
        self._policy_settings(False, policy, transaction)

    def try_unapply_policy_settings(self, policy, use_transaction=None):
        if policy not in self.get_active_policies_not_derived_from_zone():
            with self.with_transaction(use_transaction) as transaction:
                self.unapply_policy_settings(policy, transaction)

    def get_config_with_settings_dict(self, policy):
        return self.get_policy(policy).export_config_dict()

    def set_config_with_settings_dict(self, policy, settings, sender):
        # stupid wrappers to convert rich rule string to rich rule object
        def add_rule_wrapper(policy, rule_str, timeout=0, sender=None):
            self.add_rule(
                policy, Rich_Rule(rule_str=rule_str), timeout=timeout, sender=sender
            )

        def remove_rule_wrapper(policy, rule_str):
            self.remove_rule(policy, Rich_Rule(rule_str=rule_str))

        setting_to_fn = {
            "services": (self.add_service, self.remove_service),
            "ports": (self.add_port, self.remove_port),
            "icmp_blocks": (self.add_icmp_block, self.remove_icmp_block),
            "masquerade": (self.add_masquerade, self.remove_masquerade),
            "forward_ports": (self.add_forward_port, self.remove_forward_port),
            "rich_rules": (add_rule_wrapper, remove_rule_wrapper),
            "protocols": (self.add_protocol, self.remove_protocol),
            "source_ports": (self.add_source_port, self.remove_source_port),
            "ingress_zones": (self.add_ingress_zone, self.remove_ingress_zone),
            "egress_zones": (self.add_egress_zone, self.remove_egress_zone),
            "disable": (self.add_disable, self.remove_disable),
        }

        timeout = 0
        if "timeout" in settings:
            timeout = settings["timeout"]
            settings = copy.copy(settings)
            del settings["timeout"]

        # do a full config check on a temporary object before trying to make
        # the runtime changes
        old_obj = self.get_policy(policy)
        check_obj = copy.copy(old_obj)
        check_obj.import_config_dict(settings, self._fw.get_all_io_objects_dict())
        self._fw.full_check_config({"policies": [check_obj]})

        old_settings = self.get_config_with_settings_dict(policy)
        (add_settings, remove_settings) = self._fw.get_added_and_removed_settings(
            old_settings, settings
        )

        for key in remove_settings:
            if isinstance(remove_settings[key], list):
                for args in remove_settings[key]:
                    if isinstance(args, tuple):
                        setting_to_fn[key][1](policy, *args)
                    else:
                        setting_to_fn[key][1](policy, args)
            else:  # bool
                setting_to_fn[key][1](policy)

        for key in add_settings:
            if isinstance(add_settings[key], list):
                for args in add_settings[key]:
                    if isinstance(args, tuple):
                        setting_to_fn[key][0](
                            policy, *args, timeout=timeout, sender=sender
                        )
                    else:
                        setting_to_fn[key][0](
                            policy, args, timeout=timeout, sender=sender
                        )
            else:  # bool
                setting_to_fn[key][0](policy, timeout=timeout, sender=sender)

    def addTimeout(self, tag, _id):
        if _id not in self._timeouts:
            self._timeouts[_id] = tag

    def removeTimeout(self, _id):
        if _id in self._timeouts:
            GLib.source_remove(self._timeouts[_id])
            del self._timeouts[_id]

    # ingress zones

    def check_ingress_zone(self, zone):
        if not zone:
            raise FirewallError(errors.INVALID_ZONE)
        if zone not in ["HOST", "ANY"]:
            self._fw.check_zone(zone)

    def __ingress_zone_id(self, zone):
        self.check_ingress_zone(zone)
        return zone

    def add_ingress_zone(self, policy, zone, timeout=0, sender=None):
        _policy = self._fw.check_policy(policy)
        self._fw.check_timeout(timeout)
        self._fw.check_panic()
        _obj = self._policies[_policy]

        zone_id = self.__ingress_zone_id(zone)
        if zone_id in _obj.ingress_zones:
            raise FirewallError(
                errors.ALREADY_ENABLED, "'%s' already in '%s'" % (zone, _policy)
            )

        with self.with_transaction() as transaction:

            if _obj.applied:
                self._ingress_zone(True, _policy, zone, transaction)

            if timeout > 0:
                tag = GLib.timeout_add_seconds(
                    timeout, self.remove_ingress_zone, _policy, zone
                )
                self.addTimeout(tag, ("ingress_zone", _policy, zone))

            self.__register_ingress_zone(_obj, zone_id, timeout, sender)
            transaction.add_fail(self.__unregister_ingress_zone, _obj, zone_id)

            if not _obj.applied:
                self.try_apply_policy_settings(policy, use_transaction=transaction)

    def __register_ingress_zone(self, _obj, zone_id, timeout, sender):
        _obj.ingress_zones.append(zone_id)

    def remove_ingress_zone(self, policy, zone):
        _policy = self._fw.check_policy(policy)
        self._fw.check_panic()
        _obj = self._policies[_policy]

        zone_id = self.__ingress_zone_id(zone)
        if zone_id not in _obj.ingress_zones:
            raise FirewallError(
                errors.NOT_ENABLED, "'%s' not in '%s'" % (zone, _policy)
            )

        with self.with_transaction() as transaction:

            if _obj.applied:
                if len(_obj.ingress_zones) <= 1:
                    self.unapply_policy_settings(policy, transaction)
                else:
                    self._ingress_zone(False, _policy, zone, transaction)

            self.removeTimeout(("ingress_zone", _policy, zone))

            transaction.add_post(self.__unregister_ingress_zone, _obj, zone_id)

        return _policy

    def __unregister_ingress_zone(self, _obj, zone_id):
        if zone_id in _obj.ingress_zones:
            _obj.ingress_zones.remove(zone_id)

    def query_ingress_zone(self, policy, zone):
        return self.__ingress_zone_id(zone) in self.get_policy(policy).ingress_zones

    def list_ingress_zones(self, policy):
        return self.get_policy(policy).ingress_zones

    # egress zones

    def check_egress_zone(self, zone):
        if not zone:
            raise FirewallError(errors.INVALID_ZONE)
        if zone not in ["HOST", "ANY"]:
            self._fw.check_zone(zone)

    def __egress_zone_id(self, zone):
        self.check_egress_zone(zone)
        return zone

    def add_egress_zone(self, policy, zone, timeout=0, sender=None):
        _policy = self._fw.check_policy(policy)
        self._fw.check_timeout(timeout)
        self._fw.check_panic()
        _obj = self._policies[_policy]

        zone_id = self.__egress_zone_id(zone)
        if zone_id in _obj.egress_zones:
            raise FirewallError(
                errors.ALREADY_ENABLED, "'%s' already in '%s'" % (zone, _policy)
            )

        with self.with_transaction() as transaction:

            if _obj.applied:
                self._egress_zone(True, _policy, zone, transaction)

            if timeout > 0:
                tag = GLib.timeout_add_seconds(
                    timeout, self.remove_egress_zone, _policy, zone
                )
                self.addTimeout(tag, ("egress_zone", _policy, zone))

            self.__register_egress_zone(_obj, zone_id, timeout, sender)
            transaction.add_fail(self.__unregister_egress_zone, _obj, zone_id)

            if not _obj.applied:
                self.try_apply_policy_settings(policy, use_transaction=transaction)

    def __register_egress_zone(self, _obj, zone_id, timeout, sender):
        _obj.egress_zones.append(zone_id)

    def remove_egress_zone(self, policy, zone):
        _policy = self._fw.check_policy(policy)
        self._fw.check_panic()
        _obj = self._policies[_policy]

        zone_id = self.__egress_zone_id(zone)
        if zone_id not in _obj.egress_zones:
            raise FirewallError(
                errors.NOT_ENABLED, "'%s' not in '%s'" % (zone, _policy)
            )

        with self.with_transaction() as transaction:

            if _obj.applied:
                if len(_obj.egress_zones) <= 1:
                    self.unapply_policy_settings(policy, transaction)
                else:
                    self._egress_zone(False, _policy, zone, transaction)

            self.removeTimeout(("egress_zone", _policy, zone))

            transaction.add_post(self.__unregister_egress_zone, _obj, zone_id)

        return _policy

    def __unregister_egress_zone(self, _obj, zone_id):
        if zone_id in _obj.egress_zones:
            _obj.egress_zones.remove(zone_id)

    def query_egress_zone(self, policy, zone):
        return self.__egress_zone_id(zone) in self.get_policy(policy).egress_zones

    def list_egress_zones(self, policy):
        return self.get_policy(policy).egress_zones

    # RICH LANGUAGE

    def _rule_source_ipv(self, source):
        if not source:
            return None

        if source.addr:
            if checkIPnMask(source.addr):
                return "ipv4"
            elif checkIP6nMask(source.addr):
                return "ipv6"
        elif hasattr(source, "mac") and source.mac:
            return ""
        elif hasattr(source, "ipset") and source.ipset:
            self._check_ipset_type_for_source(source.ipset)
            self._check_ipset_applied(source.ipset)
            return self._ipset_family(source.ipset)

        return None

    def __rule(self, enable, policy, rule, transaction):
        self._rule_prepare(enable, policy, rule, transaction)

    def add_rule(self, policy, rule, timeout=0, sender=None):
        _policy = self._fw.check_policy(policy)
        self._fw.check_timeout(timeout)
        self._fw.check_panic()
        _obj = self._policies[_policy]

        if rule in _obj.rules:
            _name = _obj.derived_from_zone if _obj.derived_from_zone else _policy
            raise FirewallError(
                errors.ALREADY_ENABLED, "'%s' already in '%s'" % (rule, _name)
            )

        with self.with_transaction() as transaction:

            if _obj.applied:
                self.__rule(True, _policy, rule, transaction)

            if timeout > 0:
                tag = GLib.timeout_add_seconds(timeout, self.remove_rule, _policy, rule)
                self.addTimeout(tag, ("rule", _policy, rule))

            self.__register_rule(_obj, rule, timeout, sender)
            transaction.add_fail(self.__unregister_rule, _obj, rule)

        return _policy

    def __register_rule(self, _obj, rule, timeout, sender):
        _obj.rules.add(rule)

    def remove_rule(self, policy, rule):
        _policy = self._fw.check_policy(policy)
        self._fw.check_panic()
        _obj = self._policies[_policy]

        if rule not in _obj.rules:
            _name = _obj.derived_from_zone if _obj.derived_from_zone else _policy
            raise FirewallError(errors.NOT_ENABLED, "'%s' not in '%s'" % (rule, _name))

        with self.with_transaction() as transaction:

            if _obj.applied:
                self.__rule(False, _policy, rule, transaction)

            self.removeTimeout(("rule", _policy, rule))

            transaction.add_post(self.__unregister_rule, _obj, rule)

        return _policy

    def __unregister_rule(self, _obj, rule):
        if rule in _obj.rules:
            _obj.rules.remove(rule)

    def query_rule(self, policy, rule):
        return rule in self.get_policy(policy).rules

    def list_rules(self, policy):
        return [str(r) for r in self.get_policy(policy).rules]

    # SERVICES

    def check_service(self, service):
        self._fw.check_service(service)

    def __service_id(self, service):
        self.check_service(service)
        return service

    def add_service(self, policy, service, timeout=0, sender=None):
        _policy = self._fw.check_policy(policy)
        self._fw.check_timeout(timeout)
        self._fw.check_panic()
        _obj = self._policies[_policy]

        service_id = self.__service_id(service)
        if service_id in _obj.services:
            _name = _obj.derived_from_zone if _obj.derived_from_zone else _policy
            raise FirewallError(
                errors.ALREADY_ENABLED, "'%s' already in '%s'" % (service, _name)
            )

        with self.with_transaction() as transaction:

            if _obj.applied:
                self._service(True, _policy, service, transaction)

            if timeout > 0:
                tag = GLib.timeout_add_seconds(
                    timeout, self.remove_service, _policy, service
                )
                self.addTimeout(tag, ("service", _policy, service))

            self.__register_service(_obj, service_id, timeout, sender)
            transaction.add_fail(self.__unregister_service, _obj, service_id)

        return _policy

    def __register_service(self, _obj, service_id, timeout, sender):
        _obj.services.append(service_id)

    def remove_service(self, policy, service):
        _policy = self._fw.check_policy(policy)
        self._fw.check_panic()
        _obj = self._policies[_policy]

        service_id = self.__service_id(service)
        if service_id not in _obj.services:
            _name = _obj.derived_from_zone if _obj.derived_from_zone else _policy
            raise FirewallError(
                errors.NOT_ENABLED, "'%s' not in '%s'" % (service, _name)
            )

        with self.with_transaction() as transaction:

            if _obj.applied:
                self._service(False, _policy, service, transaction)

            self.removeTimeout(("service", _policy, service))

            transaction.add_post(self.__unregister_service, _obj, service_id)

        return _policy

    def __unregister_service(self, _obj, service_id):
        if service_id in _obj.services:
            _obj.services.remove(service_id)

    def query_service(self, policy, service):
        return self.__service_id(service) in self.get_policy(policy).services

    def list_services(self, policy):
        return self.get_policy(policy).services

    def get_helpers_for_service_helpers(self, helpers):
        _helpers = []
        for helper in helpers:
            try:
                _helper = self._fw.helper.get_helper(helper)
            except FirewallError:
                raise FirewallError(errors.INVALID_HELPER, helper)
            _helpers.append(_helper)
        return _helpers

    def get_helpers_for_service_modules(self, modules, enable):
        # If automatic helper assignment is turned off, helpers that
        # do not have ports defined will be replaced by the helpers
        # that the helper.module defines.
        _helpers = []
        for module in modules:
            try:
                helper = self._fw.helper.get_helper(module)
            except FirewallError:
                raise FirewallError(errors.INVALID_HELPER, module)
            if len(helper.ports) < 1:
                _module_short_name = get_nf_conntrack_short_name(helper.module)
                try:
                    _helper = self._fw.helper.get_helper(_module_short_name)
                    _helpers.append(_helper)
                except FirewallError:
                    if enable:
                        log.warning("Helper '%s' is not available" % _module_short_name)
                    continue
            else:
                _helpers.append(helper)
        return _helpers

    # PORTS

    def check_port(self, port, protocol):
        self._fw.check_port(port)
        self._fw.check_tcpudp(protocol)

    def __port_id(self, port, protocol):
        self.check_port(port, protocol)
        return (portStr(port, "-"), protocol)

    def add_port(self, policy, port, protocol, timeout=0, sender=None):
        _policy = self._fw.check_policy(policy)
        self._fw.check_timeout(timeout)
        self._fw.check_panic()
        _obj = self._policies[_policy]

        existing_port_ids = list(filter(lambda x: x[1] == protocol, _obj.ports))
        for port_id in existing_port_ids:
            if portInPortRange(port, port_id[0]):
                _name = _obj.derived_from_zone if _obj.derived_from_zone else _policy
                raise FirewallError(
                    errors.ALREADY_ENABLED,
                    "'%s:%s' already in '%s'" % (port, protocol, _name),
                )

        added_ranges, removed_ranges = coalescePortRange(
            port, [_port for (_port, _protocol) in existing_port_ids]
        )

        with self.with_transaction() as transaction:

            if _obj.applied:
                for range in added_ranges:
                    self._port(
                        True, _policy, portStr(range, "-"), protocol, transaction
                    )
                for range in removed_ranges:
                    self._port(
                        False, _policy, portStr(range, "-"), protocol, transaction
                    )

            for range in added_ranges:
                port_id = self.__port_id(range, protocol)
                self.__register_port(_obj, port_id, timeout, sender)
                transaction.add_fail(self.__unregister_port, _obj, port_id)
            for range in removed_ranges:
                port_id = self.__port_id(range, protocol)
                transaction.add_post(self.__unregister_port, _obj, port_id)

            if timeout > 0:
                tag = GLib.timeout_add_seconds(
                    timeout, self.remove_port, _policy, port, protocol
                )
                self.addTimeout(tag, ("port", _policy, port, protocol))

        return _policy

    def __register_port(self, _obj, port_id, timeout, sender):
        _obj.ports.append(port_id)

    def remove_port(self, policy, port, protocol):
        _policy = self._fw.check_policy(policy)
        self._fw.check_panic()
        _obj = self._policies[_policy]

        existing_port_ids = list(filter(lambda x: x[1] == protocol, _obj.ports))
        for port_id in existing_port_ids:
            if portInPortRange(port, port_id[0]):
                break
        else:
            _name = _obj.derived_from_zone if _obj.derived_from_zone else _policy
            raise FirewallError(
                errors.NOT_ENABLED, "'%s:%s' not in '%s'" % (port, protocol, _name)
            )

        added_ranges, removed_ranges = breakPortRange(
            port, [_port for (_port, _protocol) in existing_port_ids]
        )

        with self.with_transaction() as transaction:

            if _obj.applied:
                for range in added_ranges:
                    self._port(
                        True, _policy, portStr(range, "-"), protocol, transaction
                    )
                for range in removed_ranges:
                    self._port(
                        False, _policy, portStr(range, "-"), protocol, transaction
                    )

            for range in added_ranges:
                port_id = self.__port_id(range, protocol)
                self.__register_port(_obj, port_id, 0, None)
                transaction.add_fail(self.__unregister_port, _obj, port_id)
            for range in removed_ranges:
                port_id = self.__port_id(range, protocol)
                transaction.add_post(self.__unregister_port, _obj, port_id)

            self.removeTimeout(("port", _policy, port, protocol))

        return _policy

    def __unregister_port(self, _obj, port_id):
        if port_id in _obj.ports:
            _obj.ports.remove(port_id)

    def query_port(self, policy, port, protocol):
        for _port, _protocol in self.get_policy(policy).ports:
            if portInPortRange(port, _port) and protocol == _protocol:
                return True

        return False

    def list_ports(self, policy):
        return self.get_policy(policy).ports

    # PROTOCOLS

    def check_protocol(self, protocol):
        if not checkProtocol(protocol):
            raise FirewallError(errors.INVALID_PROTOCOL, protocol)

    def check_tcp_mss_clamp(self, tcp_mss_clamp_value):
        if not checkTcpMssClamp(tcp_mss_clamp_value):
            raise FirewallError(
                errors.INVALID_RULE,
                "tcp-mss-clamp value must be greater than or equal to 536, or the value 'pmtu'. Invalid value '%s'"
                % (tcp_mss_clamp_value),
            )

    def __protocol_id(self, protocol):
        self.check_protocol(protocol)
        return protocol

    def add_protocol(self, policy, protocol, timeout=0, sender=None):
        _policy = self._fw.check_policy(policy)
        self._fw.check_timeout(timeout)
        self._fw.check_panic()
        _obj = self._policies[_policy]

        protocol_id = self.__protocol_id(protocol)
        if protocol_id in _obj.protocols:
            _name = _obj.derived_from_zone if _obj.derived_from_zone else _policy
            raise FirewallError(
                errors.ALREADY_ENABLED, "'%s' already in '%s'" % (protocol, _name)
            )

        with self.with_transaction() as transaction:

            if _obj.applied:
                self._protocol(True, _policy, protocol, transaction)

            if timeout > 0:
                tag = GLib.timeout_add_seconds(
                    timeout, self.remove_protocol, _policy, protocol
                )
                self.addTimeout(tag, ("protocol", _policy, protocol))

            self.__register_protocol(_obj, protocol_id, timeout, sender)
            transaction.add_fail(self.__unregister_protocol, _obj, protocol_id)

        return _policy

    def __register_protocol(self, _obj, protocol_id, timeout, sender):
        _obj.protocols.append(protocol_id)

    def remove_protocol(self, policy, protocol):
        _policy = self._fw.check_policy(policy)
        self._fw.check_panic()
        _obj = self._policies[_policy]

        protocol_id = self.__protocol_id(protocol)
        if protocol_id not in _obj.protocols:
            _name = _obj.derived_from_zone if _obj.derived_from_zone else _policy
            raise FirewallError(
                errors.NOT_ENABLED, "'%s' not in '%s'" % (protocol, _name)
            )

        with self.with_transaction() as transaction:

            if _obj.applied:
                self._protocol(False, _policy, protocol, transaction)

            self.removeTimeout(("protocol", _policy, protocol))

            transaction.add_post(self.__unregister_protocol, _obj, protocol_id)

        return _policy

    def __unregister_protocol(self, _obj, protocol_id):
        if protocol_id in _obj.protocols:
            _obj.protocols.remove(protocol_id)

    def query_protocol(self, policy, protocol):
        return self.__protocol_id(protocol) in self.get_policy(policy).protocols

    def list_protocols(self, policy):
        return self.get_policy(policy).protocols

    # SOURCE PORTS

    def __source_port_id(self, port, protocol):
        self.check_port(port, protocol)
        return (portStr(port, "-"), protocol)

    def add_source_port(self, policy, port, protocol, timeout=0, sender=None):
        _policy = self._fw.check_policy(policy)
        self._fw.check_timeout(timeout)
        self._fw.check_panic()
        _obj = self._policies[_policy]

        existing_port_ids = list(filter(lambda x: x[1] == protocol, _obj.source_ports))
        for port_id in existing_port_ids:
            if portInPortRange(port, port_id[0]):
                _name = _obj.derived_from_zone if _obj.derived_from_zone else _policy
                raise FirewallError(
                    errors.ALREADY_ENABLED,
                    "'%s:%s' already in '%s'" % (port, protocol, _name),
                )

        added_ranges, removed_ranges = coalescePortRange(
            port, [_port for (_port, _protocol) in existing_port_ids]
        )

        with self.with_transaction() as transaction:

            if _obj.applied:
                for range in added_ranges:
                    self._source_port(
                        True, _policy, portStr(range, "-"), protocol, transaction
                    )
                for range in removed_ranges:
                    self._source_port(
                        False, _policy, portStr(range, "-"), protocol, transaction
                    )

            for range in added_ranges:
                port_id = self.__source_port_id(range, protocol)
                self.__register_source_port(_obj, port_id, timeout, sender)
                transaction.add_fail(self.__unregister_source_port, _obj, port_id)
            for range in removed_ranges:
                port_id = self.__source_port_id(range, protocol)
                transaction.add_post(self.__unregister_source_port, _obj, port_id)

            if timeout > 0:
                tag = GLib.timeout_add_seconds(
                    timeout, self.remove_source_port, _policy, port, protocol
                )
                self.addTimeout(tag, ("source_port", _policy, port, protocol))

        return _policy

    def __register_source_port(self, _obj, port_id, timeout, sender):
        _obj.source_ports.append(port_id)

    def remove_source_port(self, policy, port, protocol):
        _policy = self._fw.check_policy(policy)
        self._fw.check_panic()
        _obj = self._policies[_policy]

        existing_port_ids = list(filter(lambda x: x[1] == protocol, _obj.source_ports))
        for port_id in existing_port_ids:
            if portInPortRange(port, port_id[0]):
                break
        else:
            _name = _obj.derived_from_zone if _obj.derived_from_zone else _policy
            raise FirewallError(
                errors.NOT_ENABLED, "'%s:%s' not in '%s'" % (port, protocol, _name)
            )

        added_ranges, removed_ranges = breakPortRange(
            port, [_port for (_port, _protocol) in existing_port_ids]
        )

        with self.with_transaction() as transaction:

            if _obj.applied:
                for range in added_ranges:
                    self._source_port(
                        True, _policy, portStr(range, "-"), protocol, transaction
                    )
                for range in removed_ranges:
                    self._source_port(
                        False, _policy, portStr(range, "-"), protocol, transaction
                    )

            for range in added_ranges:
                port_id = self.__source_port_id(range, protocol)
                self.__register_source_port(_obj, port_id, 0, None)
                transaction.add_fail(self.__unregister_source_port, _obj, port_id)
            for range in removed_ranges:
                port_id = self.__source_port_id(range, protocol)
                transaction.add_post(self.__unregister_source_port, _obj, port_id)

            self.removeTimeout(("source_port", _policy, port, protocol))

        return _policy

    def __unregister_source_port(self, _obj, port_id):
        if port_id in _obj.source_ports:
            _obj.source_ports.remove(port_id)

    def query_source_port(self, policy, port, protocol):
        for _port, _protocol in self.get_policy(policy).source_ports:
            if portInPortRange(port, _port) and protocol == _protocol:
                return True

        return False

    def list_source_ports(self, policy):
        return self.get_policy(policy).source_ports

    # MASQUERADE

    def add_masquerade(self, policy, timeout=0, sender=None):
        _policy = self._fw.check_policy(policy)
        self._fw.check_timeout(timeout)
        self._fw.check_panic()
        _obj = self._policies[_policy]

        if _obj.masquerade:
            _name = _obj.derived_from_zone if _obj.derived_from_zone else _policy
            raise FirewallError(
                errors.ALREADY_ENABLED, "masquerade already enabled in '%s'" % _name
            )

        with self.with_transaction() as transaction:

            if _obj.applied:
                self._masquerade(True, _policy, transaction)

            if timeout > 0:
                tag = GLib.timeout_add_seconds(timeout, self.remove_masquerade, _policy)
                self.addTimeout(tag, ("masquerade", _policy))

            self.__register_masquerade(_obj, timeout, sender)
            transaction.add_fail(self.__unregister_masquerade, _obj)

        return _policy

    def __register_masquerade(self, _obj, timeout, sender):
        _obj.masquerade = True

    def remove_masquerade(self, policy):
        _policy = self._fw.check_policy(policy)
        self._fw.check_panic()
        _obj = self._policies[_policy]

        if not _obj.masquerade:
            _name = _obj.derived_from_zone if _obj.derived_from_zone else _policy
            raise FirewallError(
                errors.NOT_ENABLED, "masquerade not enabled in '%s'" % _name
            )

        with self.with_transaction() as transaction:

            if _obj.applied:
                self._masquerade(False, _policy, transaction)

            self.removeTimeout(("masquerade", _policy))

            transaction.add_post(self.__unregister_masquerade, _obj)

        return _policy

    def __unregister_masquerade(self, _obj):
        _obj.masquerade = False

    def query_masquerade(self, policy):
        return self.get_policy(policy).masquerade

    def add_disable(self, policy, timeout=0, sender=None):
        _policy = self._fw.check_policy(policy)
        self._fw.check_timeout(timeout)
        self._fw.check_panic()
        _obj = self._policies[_policy]

        if _obj.disable:
            raise FirewallError(
                errors.ALREADY_ENABLED, f"{_policy} is already disabled"
            )

        with self.with_transaction() as transaction:
            transaction.add_post(self.__register_disable, _obj, timeout, sender)

            if _obj.applied:
                self.unapply_policy_settings(_policy, transaction)

            if timeout > 0:
                tag = GLib.timeout_add_seconds(timeout, self.remove_disable, _policy)
                self.addTimeout(tag, ("disable", _policy))

        return _policy

    def __register_disable(self, _obj, timeout, sender):
        _obj.disable = True

    def remove_disable(self, policy):
        _policy = self._fw.check_policy(policy)
        self._fw.check_panic()
        _obj = self._policies[_policy]

        if not _obj.disable:
            raise FirewallError(errors.NOT_ENABLED, f"{_policy} is not disabled")

        with self.with_transaction() as transaction:
            self.__unregister_disable(_obj)
            transaction.add_fail(self.__register_disable, _obj, 0, None)

            if not _obj.applied:
                self.try_apply_policy_settings(_policy, use_transaction=transaction)

            self.removeTimeout(("disable", _policy))

        return _policy

    def __unregister_disable(self, _obj):
        _obj.disable = False

    def query_disable(self, policy):
        return self.get_policy(policy).disable

    # PORT FORWARDING

    def check_forward_port(self, ipv, port, protocol, toport=None, toaddr=None):
        self._fw.check_port(port)
        self._fw.check_tcpudp(protocol)
        if toport:
            self._fw.check_port(toport)
        if toaddr:
            if not check_single_address(ipv, toaddr):
                raise FirewallError(errors.INVALID_ADDR, toaddr)
        if not toport and not toaddr:
            raise FirewallError(
                errors.INVALID_FORWARD, "port-forwarding is missing to-port AND to-addr"
            )

    def __forward_port_id(self, port, protocol, toport=None, toaddr=None):
        if check_single_address("ipv6", toaddr):
            self.check_forward_port("ipv6", port, protocol, toport, toaddr)
        else:
            self.check_forward_port("ipv4", port, protocol, toport, toaddr)
        return (portStr(port, "-"), protocol, portStr(toport, "-"), str(toaddr))

    def add_forward_port(
        self,
        policy,
        port,
        protocol,
        toport=None,
        toaddr=None,
        timeout=0,
        sender=None,
    ):
        _policy = self._fw.check_policy(policy)
        self._fw.check_timeout(timeout)
        self._fw.check_panic()
        _obj = self._policies[_policy]

        forward_id = self.__forward_port_id(port, protocol, toport, toaddr)
        if forward_id in _obj.forward_ports:
            _name = _obj.derived_from_zone if _obj.derived_from_zone else _policy
            raise FirewallError(
                errors.ALREADY_ENABLED,
                "'%s:%s:%s:%s' already in '%s'"
                % (port, protocol, toport, toaddr, _name),
            )

        with self.with_transaction() as transaction:

            if _obj.applied:
                self._forward_port(
                    True, _policy, transaction, port, protocol, toport, toaddr
                )

            if timeout > 0:
                tag = GLib.timeout_add_seconds(
                    timeout,
                    self.remove_forward_port,
                    _policy,
                    port,
                    protocol,
                    toport,
                    toaddr,
                )
                self.addTimeout(
                    tag, ("forward_port", _policy, port, protocol, toport, toaddr)
                )

            self.__register_forward_port(_obj, forward_id, timeout, sender)
            transaction.add_fail(self.__unregister_forward_port, _obj, forward_id)

        return _policy

    def __register_forward_port(self, _obj, forward_id, timeout, sender):
        _obj.forward_ports.append(forward_id)

    def remove_forward_port(self, policy, port, protocol, toport=None, toaddr=None):
        _policy = self._fw.check_policy(policy)
        self._fw.check_panic()
        _obj = self._policies[_policy]

        forward_id = self.__forward_port_id(port, protocol, toport, toaddr)
        if forward_id not in _obj.forward_ports:
            _name = _obj.derived_from_zone if _obj.derived_from_zone else _policy
            raise FirewallError(
                errors.NOT_ENABLED,
                "'%s:%s:%s:%s' not in '%s'" % (port, protocol, toport, toaddr, _name),
            )

        with self.with_transaction() as transaction:

            if _obj.applied:
                self._forward_port(
                    False, _policy, transaction, port, protocol, toport, toaddr
                )

            self.removeTimeout(
                ("forward_port", _policy, port, protocol, toport, toaddr)
            )

            transaction.add_post(self.__unregister_forward_port, _obj, forward_id)

        return _policy

    def __unregister_forward_port(self, _obj, forward_id):
        if forward_id in _obj.forward_ports:
            _obj.forward_ports.remove(forward_id)

    def query_forward_port(self, policy, port, protocol, toport=None, toaddr=None):
        forward_id = self.__forward_port_id(port, protocol, toport, toaddr)
        return forward_id in self.get_policy(policy).forward_ports

    def list_forward_ports(self, policy):
        return self.get_policy(policy).forward_ports

    # ICMP BLOCK

    def check_icmp_block(self, icmp):
        self._fw.check_icmptype(icmp)

    def __icmp_block_id(self, icmp):
        self.check_icmp_block(icmp)
        return icmp

    def add_icmp_block(self, policy, icmp, timeout=0, sender=None):
        _policy = self._fw.check_policy(policy)
        self._fw.check_timeout(timeout)
        self._fw.check_panic()
        _obj = self._policies[_policy]

        icmp_id = self.__icmp_block_id(icmp)
        if icmp_id in _obj.icmp_blocks:
            _name = _obj.derived_from_zone if _obj.derived_from_zone else _policy
            raise FirewallError(
                errors.ALREADY_ENABLED, "'%s' already in '%s'" % (icmp, _name)
            )

        with self.with_transaction() as transaction:

            if _obj.applied:
                self._icmp_block(True, _policy, icmp, transaction)

            if timeout > 0:
                tag = GLib.timeout_add_seconds(
                    timeout, self.remove_icmp_block, _policy, icmp
                )
                self.addTimeout(tag, ("icmp_block", _policy, icmp))

            self.__register_icmp_block(_obj, icmp_id, timeout, sender)
            transaction.add_fail(self.__unregister_icmp_block, _obj, icmp_id)

        return _policy

    def __register_icmp_block(self, _obj, icmp_id, timeout, sender):
        _obj.icmp_blocks.append(icmp_id)

    def remove_icmp_block(self, policy, icmp):
        _policy = self._fw.check_policy(policy)
        self._fw.check_panic()
        _obj = self._policies[_policy]

        icmp_id = self.__icmp_block_id(icmp)
        if icmp_id not in _obj.icmp_blocks:
            _name = _obj.derived_from_zone if _obj.derived_from_zone else _policy
            raise FirewallError(errors.NOT_ENABLED, "'%s' not in '%s'" % (icmp, _name))

        with self.with_transaction() as transaction:

            if _obj.applied:
                self._icmp_block(False, _policy, icmp, transaction)

            self.removeTimeout(("icmp_block", _policy, icmp))

            transaction.add_post(self.__unregister_icmp_block, _obj, icmp_id)

        return _policy

    def __unregister_icmp_block(self, _obj, icmp_id):
        if icmp_id in _obj.icmp_blocks:
            _obj.icmp_blocks.remove(icmp_id)

    def query_icmp_block(self, policy, icmp):
        return self.__icmp_block_id(icmp) in self.get_policy(policy).icmp_blocks

    def list_icmp_blocks(self, policy):
        return self.get_policy(policy).icmp_blocks

    # ICMP BLOCK INVERSION

    def add_icmp_block_inversion(self, policy, sender=None):
        _policy = self._fw.check_policy(policy)
        self._fw.check_panic()
        _obj = self._policies[_policy]

        if _obj.icmp_block_inversion:
            _name = _obj.derived_from_zone if _obj.derived_from_zone else _policy
            raise FirewallError(
                errors.ALREADY_ENABLED,
                "icmp-block-inversion already enabled in '%s'" % _name,
            )

        with self.with_transaction() as transaction:

            if _obj.applied:
                # undo icmp blocks
                for args in _obj.icmp_blocks:
                    self._icmp_block(False, _policy, args, transaction)

                self._icmp_block_inversion(False, _policy, transaction)

            self.__register_icmp_block_inversion(_obj, sender)
            transaction.add_fail(self.__undo_icmp_block_inversion, _policy, _obj)

            # redo icmp blocks
            if _obj.applied:
                for args in _obj.icmp_blocks:
                    self._icmp_block(True, _policy, args, transaction)

                self._icmp_block_inversion(True, _policy, transaction)

        return _policy

    def __register_icmp_block_inversion(self, _obj, sender):
        _obj.icmp_block_inversion = True

    def __undo_icmp_block_inversion(self, _policy, _obj):
        with self.with_transaction() as transaction:

            # undo icmp blocks
            if _obj.applied:
                for args in _obj.icmp_blocks:
                    self._icmp_block(False, _policy, args, transaction)

            _obj.icmp_block_inversion = False

            # redo icmp blocks
            if _obj.applied:
                for args in _obj.icmp_blocks:
                    self._icmp_block(True, _policy, args, transaction)

    def remove_icmp_block_inversion(self, policy):
        _policy = self._fw.check_policy(policy)
        self._fw.check_panic()
        _obj = self._policies[_policy]

        if not _obj.icmp_block_inversion:
            _name = _obj.derived_from_zone if _obj.derived_from_zone else _policy
            raise FirewallError(
                errors.NOT_ENABLED, "icmp-block-inversion not enabled in '%s'" % _name
            )

        with self.with_transaction() as transaction:

            if _obj.applied:
                # undo icmp blocks
                for args in _obj.icmp_blocks:
                    self._icmp_block(False, _policy, args, transaction)

                self._icmp_block_inversion(False, _policy, transaction)

            self.__unregister_icmp_block_inversion(_obj)
            transaction.add_fail(self.__register_icmp_block_inversion, _obj, None)

            # redo icmp blocks
            if _obj.applied:
                for args in _obj.icmp_blocks:
                    self._icmp_block(True, _policy, args, transaction)

                self._icmp_block_inversion(True, _policy, transaction)

        return _policy

    def __unregister_icmp_block_inversion(self, _obj):
        _obj.icmp_block_inversion = False

    def query_icmp_block_inversion(self, policy):
        return self.get_policy(policy).icmp_block_inversion

    def gen_chain_rules(self, policy, create, table, chain, transaction):
        obj = self._fw.policy.get_policy(policy)
        if obj.derived_from_zone:
            # For policies derived from zones, use only the first policy in the
            # list to track chain creation. The chain names are converted to
            # zone-based names as such they're "global" for all zone derived
            # policies.
            tracking_policy = self._fw.zone._zone_policies[obj.derived_from_zone][0]
        else:
            tracking_policy = policy

        if create:
            if (
                tracking_policy in self._chains
                and (table, chain) in self._chains[tracking_policy]
            ):
                return
        else:
            if (
                tracking_policy not in self._chains
                or (table, chain) not in self._chains[tracking_policy]
            ):
                return

        for backend in self._fw.enabled_backends():
            if backend.policies_supported and table in backend.get_available_tables():
                rules = backend.build_policy_chain_rules(create, policy, table, chain)
                transaction.add_rules(backend, rules)

        self._register_chains(tracking_policy, create, [(table, chain)])
        transaction.add_fail(
            self._register_chains, tracking_policy, not create, [(table, chain)]
        )

    def _register_chains(self, policy, create, tables):
        for table, chain in tables:
            if create:
                self._chains.setdefault(policy, []).append((table, chain))
            else:
                self._chains[policy].remove((table, chain))
                if len(self._chains[policy]) == 0:
                    del self._chains[policy]

    # IPSETS

    def _ipset_family(self, name):
        if self._fw.ipset.get_type(name) == "hash:mac":
            return None
        return self._fw.ipset.get_family(name)

    def __ipset_type(self, name):
        return self._fw.ipset.get_type(name)

    def _ipset_match_flags(self, name, flag):
        return ",".join([flag] * self._fw.ipset.get_dimension(name))

    def _check_ipset_applied(self, name):
        return self._fw.ipset.check_applied(name)

    def _check_ipset_type_for_source(self, name):
        _type = self.__ipset_type(name)
        if _type not in SOURCE_IPSET_TYPES:
            raise FirewallError(
                errors.INVALID_IPSET,
                "ipset '%s' with type '%s' not usable as source" % (name, _type),
            )

    def _rule_prepare(self, enable, policy, rule, transaction, included_services=None):
        # First apply any services this service may include
        if isinstance(rule.element, Rich_Service):
            svc = self._fw.service.get_service(rule.element.name)
            if included_services is None:
                included_services = [rule.element.name]
            for include in svc.includes:
                if include in included_services:
                    continue
                self.check_service(include)
                included_services.append(include)
                self._rule_prepare(
                    enable,
                    policy,
                    dataclasses.replace(rule, element=Rich_Service(name=include)),
                    transaction,
                    included_services=included_services,
                )

        ipvs = []
        if rule.family:
            ipvs = [rule.family]
        elif rule.element and (
            isinstance(rule.element, Rich_IcmpBlock)
            or isinstance(rule.element, Rich_IcmpType)
        ):
            ict = self._fw.config.get_icmptype(rule.element.name)
            if ict.destination:
                ipvs = [ipv for ipv in ["ipv4", "ipv6"] if ipv in ict.destination]

        source_ipv = self._rule_source_ipv(rule.source)
        if source_ipv:
            if rule.family:
                # rule family is defined by user, no way to change it
                if rule.family != source_ipv:
                    raise FirewallError(
                        errors.INVALID_RULE,
                        "Source address family '%s' conflicts with rule family '%s'."
                        % (source_ipv, rule.family),
                    )
            else:
                # use the source family as rule family
                ipvs = [source_ipv]

        if not ipvs:
            ipvs = ["ipv4", "ipv6"]

        # clamp ipvs to those that are actually enabled.
        ipvs = [ipv for ipv in ipvs if self._fw.is_ipv_enabled(ipv)]

        for backend in set([self._fw.get_backend_by_ipv(x) for x in ipvs]):
            # SERVICE
            if isinstance(rule.element, Rich_Service):
                svc = self._fw.service.get_service(rule.element.name)

                destinations = []
                if len(svc.destination) > 0:
                    if rule.destination:
                        # we can not use two destinations at the same time
                        raise FirewallError(
                            errors.INVALID_RULE, "Destination conflict with service."
                        )
                    for ipv in ipvs:
                        if ipv in svc.destination and backend.is_ipv_supported(ipv):
                            destinations.append(svc.destination[ipv])
                else:
                    # dummy for the following for loop
                    destinations.append(None)

                for destination in destinations:
                    if isinstance(rule.action, Rich_Accept):
                        # only load modules for accept action
                        helpers = self.get_helpers_for_service_modules(
                            svc.modules, enable
                        )
                        helpers += self.get_helpers_for_service_helpers(svc.helpers)
                        helpers = sorted(set(helpers), key=lambda x: x.name)

                        modules = []
                        for helper in helpers:
                            module = helper.module
                            _module_short_name = get_nf_conntrack_short_name(module)
                            nat_module = module.replace("conntrack", "nat")
                            modules.append(nat_module)
                            if helper.family != "" and not backend.is_ipv_supported(
                                helper.family
                            ):
                                # no support for family ipv, continue
                                continue
                            if len(helper.ports) < 1:
                                modules.append(module)
                            else:
                                for port, proto in helper.ports:
                                    rules = backend.build_policy_helper_ports_rules(
                                        enable,
                                        policy,
                                        proto,
                                        port,
                                        destination,
                                        helper.name,
                                        _module_short_name,
                                    )
                                    transaction.add_rules(backend, rules)
                        transaction.add_modules(modules)

                    # create rules
                    for port, proto in svc.ports:
                        rules = backend.build_policy_ports_rules(
                            enable, policy, proto, port, destination, rule
                        )
                        transaction.add_rules(backend, rules)

                    for proto in svc.protocols:
                        rules = backend.build_policy_protocol_rules(
                            enable, policy, proto, destination, rule
                        )
                        transaction.add_rules(backend, rules)

                    # create rules
                    for port, proto in svc.source_ports:
                        rules = backend.build_policy_source_ports_rules(
                            enable, policy, proto, port, destination, rule
                        )
                        transaction.add_rules(backend, rules)

            # PORT
            elif isinstance(rule.element, Rich_Port):
                port = rule.element.port
                protocol = rule.element.protocol
                self.check_port(port, protocol)

                rules = backend.build_policy_ports_rules(
                    enable, policy, protocol, port, None, rule
                )
                transaction.add_rules(backend, rules)

            # PROTOCOL
            elif isinstance(rule.element, Rich_Protocol):
                protocol = rule.element.value
                self.check_protocol(protocol)

                rules = backend.build_policy_protocol_rules(
                    enable, policy, protocol, None, rule
                )
                transaction.add_rules(backend, rules)

            # TCP/MSS CLAMP
            elif isinstance(rule.element, Rich_Tcp_Mss_Clamp):
                tcp_mss_clamp_value = rule.element.value
                self.check_tcp_mss_clamp(tcp_mss_clamp_value)

                rules = backend.build_policy_tcp_mss_clamp_rules(
                    enable, policy, tcp_mss_clamp_value, None, rule
                )
                transaction.add_rules(backend, rules)

            # MASQUERADE
            elif isinstance(rule.element, Rich_Masquerade):
                if enable:
                    for ipv in ipvs:
                        if backend.is_ipv_supported(ipv):
                            transaction.add_post(enable_ip_forwarding, ipv)

                rules = backend.build_policy_masquerade_rules(enable, policy, rule)
                transaction.add_rules(backend, rules)

            # FORWARD PORT
            elif isinstance(rule.element, Rich_ForwardPort):
                port = rule.element.port
                protocol = rule.element.protocol
                toport = rule.element.to_port
                toaddr = rule.element.to_address
                for ipv in ipvs:
                    if backend.is_ipv_supported(ipv):
                        self.check_forward_port(ipv, port, protocol, toport, toaddr)
                    if toaddr and enable:
                        transaction.add_post(enable_ip_forwarding, ipv)

                rules = backend.build_policy_forward_port_rules(
                    enable, policy, port, protocol, toport, toaddr, rule
                )
                transaction.add_rules(backend, rules)

            # SOURCE PORT
            elif isinstance(rule.element, Rich_SourcePort):
                port = rule.element.port
                protocol = rule.element.protocol
                self.check_port(port, protocol)

                rules = backend.build_policy_source_ports_rules(
                    enable, policy, protocol, port, None, rule
                )
                transaction.add_rules(backend, rules)

            # ICMP BLOCK and ICMP TYPE
            elif isinstance(rule.element, Rich_IcmpBlock) or isinstance(
                rule.element, Rich_IcmpType
            ):
                ict = self._fw.config.get_icmptype(rule.element.name)

                if (
                    rule.family
                    and ict.destination
                    and rule.family not in ict.destination
                ):
                    raise FirewallError(
                        errors.INVALID_ICMPTYPE,
                        "rich rule family '%s' conflicts with icmp type '%s'"
                        % (rule.family, rule.element.name),
                    )

                if (
                    isinstance(rule.element, Rich_IcmpBlock)
                    and rule.action
                    and isinstance(rule.action, Rich_Accept)
                ):
                    # icmp block might have reject or drop action, but not accept
                    raise FirewallError(
                        errors.INVALID_RULE, "IcmpBlock not usable with accept action"
                    )

                rules = backend.build_policy_icmp_block_rules(
                    enable, policy, ict, rule, ipvs=ipvs
                )
                transaction.add_rules(backend, rules)

            elif rule.element is None:
                rules = backend.build_policy_rich_source_destination_rules(
                    enable, policy, rule
                )
                transaction.add_rules(backend, rules)

            # EVERYTHING ELSE
            else:
                raise FirewallError(
                    errors.INVALID_RULE, "Unknown element %s" % type(rule.element)
                )

    def _service(self, enable, policy, service, transaction, included_services=None):
        svc = self._fw.service.get_service(service)
        helpers = self.get_helpers_for_service_modules(svc.modules, enable)
        helpers += self.get_helpers_for_service_helpers(svc.helpers)
        helpers = sorted(set(helpers), key=lambda x: x.name)

        # First apply any services this service may include
        if included_services is None:
            included_services = [service]
        for include in svc.includes:
            if include in included_services:
                continue
            self.check_service(include)
            included_services.append(include)
            self._service(
                enable,
                policy,
                include,
                transaction,
                included_services=included_services,
            )

        # build a list of (backend, destination). The destination may be ipv4,
        # ipv6 or None
        #
        backends_ipv = []
        for ipv in ["ipv4", "ipv6"]:
            if not self._fw.is_ipv_enabled(ipv):
                continue
            backend = self._fw.get_backend_by_ipv(ipv)
            if len(svc.destination) > 0:
                if ipv in svc.destination:
                    backends_ipv.append((backend, svc.destination[ipv]))
            else:
                if (backend, None) not in backends_ipv:
                    backends_ipv.append((backend, None))

        for backend, destination in backends_ipv:
            for helper in helpers:
                module = helper.module
                _module_short_name = get_nf_conntrack_short_name(module)
                nat_module = helper.module.replace("conntrack", "nat")
                transaction.add_module(nat_module)
                if helper.family != "" and not backend.is_ipv_supported(helper.family):
                    # no support for family ipv, continue
                    continue
                if len(helper.ports) < 1:
                    transaction.add_module(module)
                else:
                    for port, proto in helper.ports:
                        rules = backend.build_policy_helper_ports_rules(
                            enable,
                            policy,
                            proto,
                            port,
                            destination,
                            helper.name,
                            _module_short_name,
                        )
                        transaction.add_rules(backend, rules)

            for port, proto in svc.ports:
                rules = backend.build_policy_ports_rules(
                    enable, policy, proto, port, destination
                )
                transaction.add_rules(backend, rules)

            for protocol in svc.protocols:
                rules = backend.build_policy_protocol_rules(
                    enable, policy, protocol, destination
                )
                transaction.add_rules(backend, rules)

            for port, proto in svc.source_ports:
                rules = backend.build_policy_source_ports_rules(
                    enable, policy, proto, port, destination
                )
                transaction.add_rules(backend, rules)

    def _port(self, enable, policy, port, protocol, transaction):
        for backend in self._fw.enabled_backends():
            if not backend.policies_supported:
                continue

            rules = backend.build_policy_ports_rules(enable, policy, protocol, port)
            transaction.add_rules(backend, rules)

    def _protocol(self, enable, policy, protocol, transaction):
        for backend in self._fw.enabled_backends():
            if not backend.policies_supported:
                continue

            rules = backend.build_policy_protocol_rules(enable, policy, protocol)
            transaction.add_rules(backend, rules)

    def _source_port(self, enable, policy, port, protocol, transaction):
        for backend in self._fw.enabled_backends():
            if not backend.policies_supported:
                continue

            rules = backend.build_policy_source_ports_rules(
                enable, policy, protocol, port
            )
            transaction.add_rules(backend, rules)

    def _masquerade(self, enable, policy, transaction):
        ipv = "ipv4"
        transaction.add_post(enable_ip_forwarding, ipv)

        backend = self._fw.get_backend_by_ipv(ipv)
        rules = backend.build_policy_masquerade_rules(enable, policy)
        transaction.add_rules(backend, rules)

    def _forward_port(
        self, enable, policy, transaction, port, protocol, toport=None, toaddr=None
    ):
        if check_single_address("ipv6", toaddr):
            ipv = "ipv6"
        else:
            ipv = "ipv4"

        if toaddr and enable:
            transaction.add_post(enable_ip_forwarding, ipv)
        backend = self._fw.get_backend_by_ipv(ipv)
        rules = backend.build_policy_forward_port_rules(
            enable, policy, port, protocol, toport, toaddr
        )
        transaction.add_rules(backend, rules)

    def _icmp_block(self, enable, policy, icmp, transaction):
        ict = self._fw.config.get_icmptype(icmp)

        ipvs = ["ipv4", "ipv6"]
        if ict.destination:
            ipvs = [ipv for ipv in ["ipv4", "ipv6"] if ipv in ict.destination]

        # clamp ipvs to those that are actually enabled.
        ipvs = [ipv for ipv in ipvs if self._fw.is_ipv_enabled(ipv)]

        for backend in set([self._fw.get_backend_by_ipv(x) for x in ipvs]):
            if not backend.policies_supported:
                continue

            rules = backend.build_policy_icmp_block_rules(
                enable, policy, ict, ipvs=ipvs
            )
            transaction.add_rules(backend, rules)

    def _icmp_block_inversion(self, enable, policy, transaction):
        target = self._policies[policy].target

        # Do not add general icmp accept rules into a trusted, block or drop
        # policy.
        if target in ["DROP", "%%REJECT%%", "REJECT"]:
            return
        if not self.query_icmp_block_inversion(policy) and target == "ACCEPT":
            # ibi target and policy target are ACCEPT, no need to add an extra
            # rule
            return

        for backend in self._fw.enabled_backends():
            if not backend.policies_supported:
                continue

            rules = backend.build_policy_icmp_block_inversion_rules(enable, policy)
            transaction.add_rules(backend, rules)

    def _ingress_egress_zone(
        self,
        enable,
        policy,
        ingress_zone,
        egress_zone,
        transaction,
    ):
        p_obj = self.get_policy(policy)

        if ingress_zone == "ANY":
            _ingress_zones = self._fw.zone.get_applied_zones()
        elif ingress_zone == "HOST":
            _ingress_zones = [ingress_zone]
        elif self._fw.zone.get_zone(ingress_zone).applied:
            _ingress_zones = [ingress_zone]
        else:
            _ingress_zones = []

        if egress_zone == "ANY":
            _egress_zones = self._fw.zone.get_applied_zones()
        elif egress_zone == "HOST":
            _egress_zones = [egress_zone]
        elif self._fw.zone.get_zone(egress_zone).applied:
            _egress_zones = [egress_zone]
        else:
            _egress_zones = []

        for _ingress_zone in _ingress_zones:
            for _egress_zone in _egress_zones:
                for backend in self._fw.enabled_backends():
                    if not backend.policies_supported:
                        continue

                    # For zone derived policies, we'll have two: (zone, "ANY")
                    # and ("ANY", zone). This results in rule duplication as
                    # ingress-zone == egress-zone is covered by both of these.
                    #
                    # Skip one of them.
                    if (
                        p_obj.derived_from_zone
                        and _ingress_zone == _egress_zone
                        and ingress_zone == "ANY"
                    ):
                        continue

                    for table, chain in self._get_table_chains_for_policy_dispatch(
                        policy
                    ):
                        # The backend wants the _target_ policy to jump to.
                        #
                        # Zone derived policies have unique conventions due to
                        # $HISTORY. As such, the policy passed to the backends
                        # sometimes needs to represent the ingress zone (PREROUTING,
                        # FORWARD) and sometimes the egress zone (POSTROUTING).
                        # For example, the policy passed to this function may
                        # be derived from "internal" zone where
                        # ingress-zones=ANY and egress-zones=internal. However,
                        # the backends need to know the _target_ policy.
                        #
                        # Regular policies don't have this issue.
                        #
                        # See add_zone()
                        target_policy = policy
                        if p_obj.derived_from_zone:
                            if chain in ["FORWARD", "PREROUTING"]:
                                target_policy = self._fw.zone.policy_name_from_zones(
                                    _ingress_zone, "ANY"
                                )
                            elif chain == "POSTROUTING":
                                target_policy = self._fw.zone.policy_name_from_zones(
                                    "ANY", _egress_zone
                                )

                        rules = backend.build_policy_dispatch_rules(
                            enable,
                            _ingress_zone,
                            _egress_zone,
                            target_policy,
                            table,
                            chain,
                        )
                        transaction.add_rules(backend, rules)

    def _ingress_zone(
        self,
        enable,
        policy,
        ingress_zone,
        transaction,
    ):
        for egress_zone in self.list_egress_zones(policy):
            self._ingress_egress_zone(
                enable,
                policy,
                ingress_zone,
                egress_zone,
                transaction,
            )

    def _egress_zone(
        self,
        enable,
        policy,
        egress_zone,
        transaction,
    ):
        for ingress_zone in self.list_ingress_zones(policy):
            self._ingress_egress_zone(
                enable,
                policy,
                ingress_zone,
                egress_zone,
                transaction,
            )

    def _get_table_chains_for_policy_dispatch(self, policy):
        """Create a list of (table, chain) needed for policy dispatch"""
        obj = self._policies[policy]
        if "HOST" in obj.egress_zones:
            # zone --> HOST
            tc = [("filter", "INPUT"), ("nat", "PREROUTING"), ("mangle", "PREROUTING")]
            # iptables backend needs to put conntrack helper rules in raw
            # prerouting.
            if not self._fw.nftables_enabled:
                tc.append(("raw", "PREROUTING"))
            return tc
        elif "HOST" in obj.ingress_zones:
            # HOST --> zone/any
            return [("filter", "OUTPUT"), ("nat", "OUTPUT")]
        else:
            # ANY/zone -> ANY/zone
            tc = [
                ("filter", "FORWARD"),
                ("nat", "PREROUTING"),
                ("nat", "POSTROUTING"),
                ("mangle", "PREROUTING"),
            ]
            # iptables backend needs to put conntrack helper rules in raw
            # prerouting.
            if not self._fw.nftables_enabled:
                tc.append(("raw", "PREROUTING"))
            return tc

    def _get_table_chains_for_zone_dispatch(self, ingress_zone, egress_zone):
        if "HOST" == ingress_zone:
            return [("filter", "OUTPUT"), ("nat", "OUTPUT")]
        elif "HOST" == egress_zone:
            tc = [("filter", "INPUT"), ("nat", "PREROUTING"), ("mangle", "PREROUTING")]
            # iptables backend needs to put conntrack helper rules in raw
            # prerouting.
            if not self._fw.nftables_enabled:
                tc.append(("raw", "PREROUTING"))
            return tc
        else:  # ANY/zone -> ANY/zone
            tc = [
                ("filter", "FORWARD"),
                ("nat", "PREROUTING"),
                ("nat", "POSTROUTING"),
                ("mangle", "PREROUTING"),
            ]
            # iptables backend needs to put conntrack helper rules in raw
            # prerouting.
            if not self._fw.nftables_enabled:
                tc.append(("raw", "PREROUTING"))

            return tc

    def policy_base_chain_name(self, policy, table, policy_prefix, isSNAT=False):
        obj = self._fw.policy.get_policy(policy)
        if obj.derived_from_zone:
            suffix = obj.derived_from_zone
        else:
            suffix = policy_prefix + policy

        if "HOST" in obj.egress_zones:
            # zone/any --> Host
            if table == "filter":
                return "IN_" + suffix
            elif table == "raw":
                # NOTE: nftables doesn't actually use this. Only iptables
                return "PRE_" + suffix
            elif table in ["mangle", "nat"]:
                return "PRE_" + suffix
        elif "HOST" in obj.ingress_zones:
            # HOST --> zone/any
            if table in ["filter", "nat"]:
                return "OUT_" + suffix
        elif "ANY" in obj.egress_zones:
            # zone/any --> any
            if table == "filter":
                return "FWD_" + suffix
            elif table == "nat":
                if isSNAT:
                    return "POST_" + suffix
                else:
                    return "PRE_" + suffix
            elif table in ["mangle", "raw"]:
                return "PRE_" + suffix
        elif "ANY" in obj.ingress_zones:
            # any --> zone
            if table == "filter":
                return "FWD_" + suffix
            elif table == "nat":
                if isSNAT:
                    return "POST_" + suffix
                else:
                    return "PRE_" + suffix
            elif table in ["mangle", "raw"]:
                return "PRE_" + suffix
        elif not obj.derived_from_zone:
            # zone --> zone
            if table == "filter":
                return "FWD_" + suffix
            elif table == "nat":
                if isSNAT:
                    return "POST_" + suffix
                else:
                    return "PRE_" + suffix
            elif table in ["mangle", "raw"]:
                return "PRE_" + suffix
        raise FirewallError(
            errors.INVALID_POLICY,
            "Can't convert policy to chain name: %s, %s, %s" % (policy, table, isSNAT),
        )
